from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import subprocess
import numpy as np
import pyopencl as cl
import pyopencl.array  # NOQA
from functools import partial


def drive_test_completeness(ctx, queue, dim, q_order):

    n_levels = 2  # 2^(n_levels-1) subintervals in 1D, must be at least 2

    # bounding box
    a = -1
    b = 1

    dtype = np.float64

    def source_field(x):
        assert len(x) == dim
        return 1

    # {{{ generate quad points

    import volumential.meshgen as mg

    q_points, q_weights, q_radii = mg.make_uniform_cubic_grid(
        degree=q_order, level=n_levels, dim=dim
    )

    assert len(q_points) == len(q_weights)
    assert q_points.shape[1] == dim

    q_points_org = q_points
    q_points = np.ascontiguousarray(np.transpose(q_points))

    from pytools.obj_array import make_obj_array

    q_points = make_obj_array(
        [cl.array.to_device(queue, q_points[i]) for i in range(dim)]
    )

    q_weights = cl.array.to_device(queue, q_weights)
    if q_radii is not None:
        q_radii = cl.array.to_device(queue, q_radii)

    # }}}

    # {{{ discretize the source field

    source_vals = cl.array.to_device(
        queue, np.array([source_field(qp) for qp in q_points_org])
    )

    # }}} End discretize the source field

    # {{{ build tree and traversals

    from boxtree.tools import AXIS_NAMES

    axis_names = AXIS_NAMES[:dim]

    from pytools import single_valued

    coord_dtype = single_valued(coord.dtype for coord in q_points)
    from boxtree.bounding_box import make_bounding_box_dtype

    bbox_type, _ = make_bounding_box_dtype(ctx.devices[0], dim, coord_dtype)

    bbox = np.empty(1, bbox_type)
    for ax in axis_names:
        bbox["min_" + ax] = a
        bbox["max_" + ax] = b

    # tune max_particles_in_box to reconstruct the mesh
    # TODO: use points from FieldPlotter are used as target points for better
    # visuals
    from boxtree import TreeBuilder

    tb = TreeBuilder(ctx)
    tree, _ = tb(
        queue,
        particles=q_points,
        targets=q_points,
        bbox=bbox,
        max_particles_in_box=q_order ** dim * (2 ** dim) - 1,
        kind="adaptive-level-restricted",
    )

    from boxtree.traversal import FMMTraversalBuilder

    tg = FMMTraversalBuilder(ctx)
    trav, _ = tg(queue, tree)

    # }}} End build tree and traversals

    from volumential.table_manager import NearFieldInteractionTableManager

    subprocess.check_call(['rm', '-f', 'nft.hdf5'])

    with NearFieldInteractionTableManager("nft.hdf5") as tm:

        nft, _ = tm.get_table(dim, "Constant", q_order,
                queue=queue, n_levels=1, alpha=0,
                compute_method="DrosteSum",
                n_brick_quad_points=50,
                adaptive_level=False, use_symmetry=True)

    subprocess.check_call(['rm', 'nft.hdf5'])

    # {{{ expansion wrangler

    from sumpy.kernel import LaplaceKernel
    from sumpy.expansion.multipole import VolumeTaylorMultipoleExpansion
    from sumpy.expansion.local import VolumeTaylorLocalExpansion

    knl = LaplaceKernel(dim)
    out_kernels = [knl]
    local_expn_class = VolumeTaylorLocalExpansion
    mpole_expn_class = VolumeTaylorMultipoleExpansion

    from volumential.expansion_wrangler_fpnd import (
            FPNDExpansionWranglerCodeContainer,
            FPNDExpansionWrangler)

    wcc = FPNDExpansionWranglerCodeContainer(
        ctx,
        partial(mpole_expn_class, knl),
        partial(local_expn_class, knl),
        out_kernels,
        exclude_self=True,
    )

    wrangler = FPNDExpansionWrangler(
        code_container=wcc,
        queue=queue,
        tree=tree,
        near_field_table=nft,
        dtype=dtype,
        fmm_level_to_order=lambda kernel, kernel_args, tree, lev: 1,
        quad_order=q_order,
    )

    # }}} End sumpy expansion for laplace kernel
    pot = wrangler.eval_direct(
        trav.target_boxes,
        trav.neighbor_source_boxes_starts,
        trav.neighbor_source_boxes_lists,
        mode_coefs=source_vals,
    )
    pot = pot[0]
    for p in pot[0]:
        assert(abs(p - 2**dim) < 1e-8)


def test_completeness_1(ctx_factory):

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    drive_test_completeness(ctx, queue, 2, 1)
    drive_test_completeness(ctx, queue, 3, 1)


def test_completeness(longrun, ctx_factory):

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    for q in range(2, 4):
        drive_test_completeness(ctx, queue, 2, q)
        drive_test_completeness(ctx, queue, 3, q)



from __future__ import division, absolute_import, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# import volumential.volume_fmm as vfmm

import numpy as np
import pyopencl as cl

# import pyopencl.array
# import boxtree as bt
# import sumpy as sp
# import volumential as vm

from functools import partial

import logging

logger = logging.getLogger(__name__)

import volumential.meshgen as mg

import pytest

# {{{ make sure context getter works


def test_cl_ctx_getter(ctx_factory):
    logging.basicConfig(level=logging.INFO)

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    a_np = np.random.rand(50000).astype(np.float32)
    b_np = np.random.rand(50000).astype(np.float32)

    mf = cl.mem_flags
    a_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a_np)
    b_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b_np)

    prg = cl.Program(
        ctx,
        """
    __kernel void sum(
        __global const float *a_g, __global const float *b_g,
        __global float *res_g) {
      int gid = get_global_id(0);
      res_g[gid] = a_g[gid] + b_g[gid];
    }
    """,
    ).build()

    res_g = cl.Buffer(ctx, mf.WRITE_ONLY, a_np.nbytes)
    prg.sum(queue, a_np.shape, None, a_g, b_g, res_g)

    res_np = np.empty_like(a_np)
    cl.enqueue_copy(queue, res_np, res_g)

    # Check on CPU with Numpy:
    assert np.linalg.norm(res_np - (a_np + b_np)) < 1e-12


# }}}

# {{{ laplace volume potential


@pytest.fixture
def laplace_problem(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    dim = 2
    dtype = np.float64

    q_order = 2  # quadrature order
    n_levels = 3  # 2^(n_levels-1) subintervals in 1D

    # adaptive_mesh = True
    n_refinement_loops = 100
    refined_n_cells = 1000
    rratio_top = 0.2
    rratio_bot = 0.5

    # bounding box
    a = -1
    b = 1

    m_order = 15  # multipole order

    alpha = 160 / np.sqrt(2)

    def source_field(x):
        assert len(x) == dim
        assert dim == 2
        norm2 = x[0] ** 2 + x[1] ** 2
        lap_u = (4 * alpha ** 2 * norm2 - 4 * alpha) * np.exp(-alpha * norm2)
        return -lap_u

    def exact_solu(x, y):
        norm2 = x ** 2 + y ** 2
        return np.exp(-alpha * norm2)

    # {{{ generate quad points

    mesh = mg.MeshGen2D(q_order, n_levels)
    iloop = 0
    while mesh.n_active_cells() < refined_n_cells:
        iloop += 1
        crtr = np.array(
            [
                np.abs(source_field(c) * m)
                for (c, m) in zip(mesh.get_cell_centers(), mesh.get_cell_measures())
            ]
        )
        mesh.update_mesh(crtr, rratio_top, rratio_bot)
        if iloop > n_refinement_loops:
            print("Max number of refinement loops reached.")
            break

    q_points = mesh.get_q_points()
    q_weights = mesh.get_q_weights()
    # q_radii = None

    assert len(q_points) == len(q_weights)
    assert q_points.shape[1] == dim

    q_points_org = q_points
    q_points = np.ascontiguousarray(np.transpose(q_points))

    from pytools.obj_array import make_obj_array

    q_points = make_obj_array(
        [cl.array.to_device(queue, q_points[i]) for i in range(dim)]
    )

    q_weights = cl.array.to_device(queue, q_weights)
    # q_radii = cl.array.to_device(queue, q_radii)

    # }}}

    # {{{ discretize the source field

    source_vals = cl.array.to_device(
        queue, np.array([source_field(qp) for qp in q_points_org])
    )

    # particle_weigt = source_val * q_weight

    # }}} End discretize the source field

    # {{{ build tree and traversals

    from boxtree.tools import AXIS_NAMES

    axis_names = AXIS_NAMES[:dim]

    from pytools import single_valued

    coord_dtype = single_valued(coord.dtype for coord in q_points)
    from boxtree.bounding_box import make_bounding_box_dtype

    bbox_type, _ = make_bounding_box_dtype(ctx.devices[0], dim, coord_dtype)

    bbox = np.empty(1, bbox_type)
    for ax in axis_names:
        bbox["min_" + ax] = a
        bbox["max_" + ax] = b

    # tune max_particles_in_box to reconstruct the mesh
    from boxtree import TreeBuilder

    tb = TreeBuilder(ctx)
    tree, _ = tb(
        queue,
        particles=q_points,
        targets=q_points,
        bbox=bbox,
        max_particles_in_box=q_order ** 2 * 4 - 1,
        kind="adaptive-level-restricted",
    )

    from boxtree.traversal import FMMTraversalBuilder

    tg = FMMTraversalBuilder(ctx)
    trav, _ = tg(queue, tree)

    # }}} End build tree and traversals

    # {{{ build near field potential table

    from volumential.table_manager import NearFieldInteractionTableManager

    tm = NearFieldInteractionTableManager("../nft.hdf5")
    nftable, _ = tm.get_table(dim, "Laplace", q_order)

    # }}} End build near field potential table

    # {{{ sumpy expansion for laplace kernel

    from sumpy.kernel import LaplaceKernel

    # from sumpy.expansion.multipole import VolumeTaylorMultipoleExpansion
    # from sumpy.expansion.local import VolumeTaylorLocalExpansion

    from sumpy.expansion.multipole import (
        LaplaceConformingVolumeTaylorMultipoleExpansion,
    )
    from sumpy.expansion.local import LaplaceConformingVolumeTaylorLocalExpansion

    knl = LaplaceKernel(dim)
    out_kernels = [knl]
    local_expn_class = LaplaceConformingVolumeTaylorLocalExpansion
    mpole_expn_class = LaplaceConformingVolumeTaylorMultipoleExpansion
    # local_expn_class = VolumeTaylorLocalExpansion
    # mpole_expn_class = VolumeTaylorMultipoleExpansion

    exclude_self = True
    from volumential.expansion_wrangler_fpnd import (
            FPNDExpansionWranglerCodeContainer,
            FPNDExpansionWrangler)

    wcc = FPNDExpansionWranglerCodeContainer(
        ctx,
        partial(mpole_expn_class, knl),
        partial(local_expn_class, knl),
        out_kernels,
        exclude_self=exclude_self,
    )

    if exclude_self:
        target_to_source = np.arange(tree.ntargets, dtype=np.int32)
        self_extra_kwargs = {"target_to_source": target_to_source}
    else:
        self_extra_kwargs = {}

    wrangler = FPNDExpansionWrangler(
        code_container=wcc,
        queue=queue,
        tree=tree,
        near_field_table=nftable,
        dtype=dtype,
        fmm_level_to_order=lambda kernel, kernel_args, tree, lev: m_order,
        quad_order=q_order,
        self_extra_kwargs=self_extra_kwargs,
    )

    # }}} End sumpy expansion for laplace kernel

    return trav, wrangler, source_vals, q_weights


# }}} End laplace volume potential


@pytest.mark.skipif(
    mg.provider != "meshgen_dealii", reason="Adaptive mesh module is not available"
)
def test_volume_fmm_laplace(laplace_problem):

    trav, wrangler, source_vals, q_weights = laplace_problem

    from volumential.volume_fmm import drive_volume_fmm

    direct_pot, = drive_volume_fmm(
        trav, wrangler, source_vals * q_weights, source_vals, direct_evaluation=True
    )

    fmm_pot, = drive_volume_fmm(
        trav, wrangler, source_vals * q_weights, source_vals, direct_evaluation=False
    )

    assert max(abs(direct_pot - fmm_pot)) < 1e-6


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        pytest.main([__file__])


# vim: filetype=pyopencl:foldmethod=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

logger = logging.getLogger(__name__)

from abc import ABCMeta, abstractmethod

__doc__ = """
.. autoclass::ExpansionWranglerInterface
"""

# {{{ expansion wrangler interface

# NOTE: abstractmethod's signatures (arguement lists) are not enforced


class ExpansionWranglerInterface(object):
    """
        Abstract expansion handling interface.
        The interface is adapted from, and stays compatible with boxtree/fmm.
        TODO: Update docstrings
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def multipole_expansion_zeros(self):
        """
        Construct arrays to store multipole expansions for all boxes
        """
        pass

    @abstractmethod
    def local_expansion_zeros(self):
        """
        Construct arrays to store multipole expansions for all boxes
        """
        pass

    @abstractmethod
    def output_zeros(self):
        """
        Construct arrays to store potential values for all target points
        """
        pass

    @abstractmethod
    def reorder_sources(self, source_array):
        """
        Return a copy of *source_array* in tree source order.
        *source_array* is in user source order.
        """
        pass

    @abstractmethod
    def reorder_targets(self, source_array):
        """
        Return a copy of *target_array* in tree source order.
        *target_array* is in user target order.
        """
        pass

    @abstractmethod
    def reorder_potentials(self, potentials):
        """
        Return a copy of *potentials* in user target order.
        *source_weights* is in tree target order.
        """
        pass

    @abstractmethod
    def form_multipoles(self, level_start_source_box_nrs, source_boxes, src_weights):
        """
        Return an expansions array containing multipole expansions
        in *source_boxes* due to sources with *src_weights*.
        """
        pass

    @abstractmethod
    def coarsen_multipoles(
        self, level_start_source_parent_box_nrs, source_parent_boxes, mpoles
    ):
        """
        For each box in *source_parent_boxes*, gather (and translate)
        the box's children's multipole expansions in *mpoles* and add
        the resulting expansion into the box's multipole expansion
        in *mpoles*.

        :returns: *mpoles*
        """
        pass

    @abstractmethod
    def eval_direct(
        self, target_boxes, neighbor_sources_starts, neighbor_sources_lists
    ):
        """
        For each box in *target_boxes*, evaluate the influence of the
        neighbor sources due to *src_weights*

        This step amounts to looking up the corresponding entries in a
        pre-built table.

        :returns: a new potential array, see :meth:`output_zeros`.
        """
        pass

    @abstractmethod
    def multipole_to_local(
        self,
        level_start_target_or_target_parent_box_nrs,
        target_or_target_parent_boxes,
        starts,
        lists,
        mpole_exps,
    ):
        """
        For each box in *target_or_target_parent_boxes*, translate and add
        the influence of the multipole expansion in *mpole_exps* into a new
        array of local expansions.

        :returns: a new (local) expansion array.
        """
        pass

    @abstractmethod
    def eval_multipoles(
        self, level_start_target_box_nrs, target_boxes, starts, lists, mpole_exps
    ):
        """
        For each box in *target_boxes*, evaluate the multipole expansion in
        *mpole_exps* in the nearby boxes given in *starts* and *lists*, and
        return a new potential array.

        :returns: a new potential array, see :meth:`output_zeros`.
        """
        pass

    @abstractmethod
    def form_locals(
        self,
        level_start_target_or_target_parent_box_nrs,
        target_or_target_parent_boxes,
        starts,
        lists,
        src_weights,
    ):
        """
        For each box in *target_or_target_parent_boxes*, form local
        expansions due to the sources in the nearby boxes given in *starts* and
        *lists*, and return a new local expansion array.

        :returns: a new local expansion array
        """
        pass

    @abstractmethod
    def refine_locals(
        self,
        level_start_target_or_target_parent_box_nrs,
        target_or_target_parent_boxes,
        local_exps,
    ):
        """
        For each box in *child_boxes*,
        translate the box's parent's local expansion in *local_exps* and add
        the resulting expansion into the box's local expansion in *local_exps*.

        :returns: *local_exps*
        """
        pass

    @abstractmethod
    def eval_locals(self, level_start_target_box_nrs, target_boxes, local_exps):
        """For each box in *target_boxes*, evaluate the local expansion in
        *local_exps* and return a new potential array.

        :returns: a new potential array, see :meth:`output_zeros`.
        """
        pass

    @abstractmethod
    def finalize_potentials(self, potentials):
        """
        Postprocess the reordered potentials. This is where global scaling
        factors could be applied.
        """
        pass


# }}} End expansion wrangler interface

# {{{ code container interface


class ExpansionWranglerCodeContainerInterface(object):
    """
        Abstract expansion code container interface.
        The interface is adapted from, and stays compatible with boxtree/fmm.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def get_wrangler(self, *args, **kwargs):
        """Makes a wrangler object.
        """
        pass

# }}} End code container interface

# vim: filetype=pyopencl:fdm=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2019 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import pyopencl as cl
from pytools.obj_array import make_obj_array
from boxtree.pyfmmlib_integration import FMMLibRotationData

# {{{ bounding box factory


class BoundingBoxFactory():
    def __init__(self, dim,
            center=None, radius=None,
            dtype=np.float64):
        self.dim = dim
        self.dtype = dtype

        self.center = center
        self.radius = radius

        self.box_center = None
        self.box_radius = None
        self.lbounds = None
        self.ubounds = None

    @property
    def materialized(self):
        """Returns true if the bounding box is readily determined,
        i.e., the __call__() method has been called.
        """
        return (self.box_center is not None)

    def __call__(self, ctx,
            expand_to_hold_mesh=None, mesh_padding_factor=0.05):
        """If expand the box to be large enough to enclose a
        mesh object.

        When center is given during initialization, the box center is
        preserved during expansion.

        When radius is given during initialization, the radius will not
        shrink and will be at least the specified value.

        :arg ctx: OpenCL context.

        :arg expand_to_hold_mesh: A :class:`meshmode.mesh.Mesh` object.

        :arg mesh_padding_factor: Controls how far out the box boundary is
            relative to the mesh.
        """
        if not expand_to_hold_mesh:
            if (self.center is None) or (self.radius is None):
                raise ValueError(
                        "A bounding box without presets "
                        "must be adapted to an mesh.")
            box_center = self.center
            box_radius = self.radius
        else:
            a = np.min(expand_to_hold_mesh.vertices, axis=0)
            b = np.max(expand_to_hold_mesh.vertices, axis=0)
            c = (a + b) * 0.5

            if self.center is None:
                box_center = c
            else:
                box_center = self.center

            if self.radius is None:
                box_radius = np.max(b - c)
            else:
                box_radius = max(self.radius, np.max(b - c))

        self.box_center = box_center
        self.box_radius = box_radius
        self.lbounds = box_center - box_radius
        self.ubounds = box_center + box_radius

        from boxtree.tools import AXIS_NAMES
        from boxtree.bounding_box import make_bounding_box_dtype

        axis_names = AXIS_NAMES[:self.dim]
        bbox_type, _ = make_bounding_box_dtype(
                ctx.devices[0], self.dim, self.dtype)

        bbox = np.empty(1, bbox_type)
        for ax, iaxis in zip(axis_names, range(self.dim)):
            bbox["min_" + ax] = self.lbounds[iaxis]
            bbox["max_" + ax] = self.ubounds[iaxis]

        return bbox

# }}} End bounding box factory

# {{{ box mesh data factory


class BoxFMMGeometryFactory():
    """Builds the geometry information needed for performing volume FMM.
    The objects of this class is designed to be "interactive", i.e., be
    manipulated repeatedly and on demand, produce a series of desired
    :class:`BoxFMMGeometryData` objects.

    .. attribute:: cl_context
        A :class:`pyopencl.Context` object.

    .. attribute:: quadrature_formula
        The quadrature formula, a :class:`modepy.quadrature.Quadrature` object.

    .. attribute:: order
        The quadrature order.

    .. attribute:: nlevels
        The number of (initial) levels.

    .. attribute:: bbox_getter
        A :class:`volumential.geometry.BoundingBoxFactory` object.

    .. attribute:: engine
        Box mesh generator that implements the
        :class:`volumential.meshgen.MeshGenBase` interface.
    """
    def __init__(
            self, cl_ctx, dim, quadrature_formula,
            order, nlevels, bbox_getter,
            expand_to_hold_mesh=None, mesh_padding_factor=0.05):
        """
        :arg bbox_getter: A :class:`BoundingBoxFactory` object.

        :arg expand_to_hold_mesh: (Optional) A :class:`meshmode.mesh.Mesh`
                                  object.

        :arg mesh_padding_factor: (Optional) Controls how far out the box
                                  boundary is relative to the mesh.
        """
        if dim == 1:
            raise NotImplementedError()
        elif dim == 2:
            from volumential.meshgen import MeshGen2D
            self._engine_class = MeshGen2D
        elif dim == 3:
            from volumential.meshgen import MeshGen3D
            self._engine_class = MeshGen3D

        self.dim = dim
        self.quadrature_formula = quadrature_formula

        self.bbox_getter = bbox_getter

        self.cl_context = cl_ctx
        self._bbox = bbox_getter(self.cl_context,
                expand_to_hold_mesh, mesh_padding_factor)

        self.order = order
        self.nlevels = nlevels

        # the engine generates meshes centered at the origin
        a = -1 * self.bbox_getter.radius
        b = self.bbox_getter.radius
        self.engine = self._engine_class(self.order, self.nlevels, a, b)

    def reinit(
            self, order=None, nlevels=None,
            quadrature_formula=None,
            expand_to_hold_mesh=None, mesh_padding_factor=None):
        """Resets the engine to its initial state, and optionally also
        changes some parameters.
        """
        if order is not None:
            self.order = order

        if nlevels is not None:
            self.nlevels = nlevels

        if expand_to_hold_mesh is not None:
            self._bbox = self.bbox_getter(self.cl_context,
                    expand_to_hold_mesh, mesh_padding_factor)

        # the engine generates meshes centered at the origin
        a = -1 * self.bbox_getter.radius
        b = self.bbox_getter.radius
        self.engine = self._engine_class(order, nlevels, a, b)

    def _get_q_points(self, queue=None):
        q_points_pre = self.engine.get_q_points()
        q_points = np.ascontiguousarray(np.transpose(q_points_pre))

        # re-center the box
        q_points = q_points + self.bbox_getter.box_center[:, None]

        if queue is None:
            return q_points
        else:
            return make_obj_array(
                    [cl.array.to_device(queue, q_points[i])
                     for i in range(self.dim)])

    def _get_q_weights(self, queue=None):
        q_weights = self.engine.get_q_weights()
        if queue is None:
            return q_weights
        else:
            return cl.array.to_device(queue, q_weights)

    def _get_active_cell_centers(self, queue=None):
        cell_centers_pre = self.engine.get_cell_centers()
        cell_centers = np.ascontiguousarray(np.transpose(cell_centers_pre))

        # re-center the box
        cell_centers = cell_centers + self.bbox_getter.box_center[:, None]

        if queue is None:
            return cell_centers
        else:
            return make_obj_array(
                    [cl.array.to_device(queue, cell_centers[i])
                     for i in range(self.dim)])

    def _get_active_cell_measures(self, queue=None):
        cell_measures = self.engine.get_cell_measures()
        if queue is None:
            return cell_measures
        else:
            return cl.array.to_device(queue, cell_measures)

    def _get_active_cell_extents(self, queue=None):
        return self._get_active_cell_measures(queue) ** (1 / self.dim)

    @property
    def n_cells(self):
        return self.engine.n_cells()

    @property
    def n_active_cells(self):
        return self.engine.n_active_cells()

    @property
    def n_q_points_per_cell(self):
        return len(self.quadrature_formula.weights)

    def __call__(self, queue=None):
        if queue is None:
            queue = cl.CommandQueue(self.cl_context)

        from boxtree import TreeBuilder
        tb = TreeBuilder(self.cl_context)

        q_points = self._get_q_points(queue)

        tree, _ = tb(queue,
                particles=q_points, targets=q_points, bbox=self._bbox,
                max_particles_in_box=(
                    self.n_q_points_per_cell * (2**self.dim) - 1),
                kind="adaptive-level-restricted")

        from boxtree.traversal import FMMTraversalBuilder
        tg = FMMTraversalBuilder(self.cl_context)
        trav, _ = tg(queue, tree)

        return BoxFMMGeometryData(
                self.cl_context,
                q_points, self._get_q_weights(queue),
                tree, trav)

# }}} End box mesh data factory


class BoxFMMGeometryData(FMMLibRotationData):
    """A container class for geometric data associated with volume
    FMM evaluations. The objects of this class are treated as immutable.

    .. attribute:: nodes
    .. attribute:: weights
    .. attribute:: tree
    .. attribute:: traversal
    """
    def __init__(self, cl_context, q_points, q_weights, tree, trav):
        self.cl_context = cl_context
        self.queue = cl.CommandQueue(cl_context)

        self.nodes = q_points
        self.weights = q_weights
        self.tree = tree
        self.traversal = trav

        super().__init__(self.queue, self.traversal)



from __future__ import absolute_import, division, print_function
import six

__copyright__ = "Copyright (C) 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import loopy as lp
import pyopencl as cl
import pymbolic as pmbl
from pytools import memoize_method
from pymbolic.primitives import Variable as VariableType
from pymbolic.primitives import Expression as ExpressionType

import logging

logger = logging.getLogger(__name__)

# {{{ clean files


def clean_file(filename, new_name=None):
    """Remove/rename file if exists.
    Fails silently when the file does not exist.
    Useful for, for example, writing output files that
    are meant to overwrite existing ones.
    """
    import os

    if new_name is None:
        try:
            os.remove(filename)
        except OSError:
            pass
    else:
        try:
            os.rename(filename, new_name)
        except OSError:
            pass

# }}} End clean files

# {{{ loopy kernel cache wrapper


class KernelCacheWrapper(object):
    # FIXME: largely code duplication with sumpy.

    def __init__(self):
        self.name = "KernelCacheWrapper"
        raise RuntimeError(
                "KernelCacheWrapper objects should not be constructed")

    def get_cache_key(self):
        raise NotImplementedError("Unimplemented cache key")

    def get_kernel(self):
        raise NotImplementedError()

    def get_optimized_kernel(self):
        raise NotImplementedError()

    @memoize_method
    def get_cached_optimized_kernel(self, **kwargs):
        from sumpy import code_cache, CACHING_ENABLED, OPT_ENABLED

        if CACHING_ENABLED:
            import loopy.version
            from sumpy.version import KERNEL_VERSION as SUMPY_KERNEL_VERSION
            from volumential.version import KERNEL_VERSION
            cache_key = (
                    self.get_cache_key()
                    + tuple(sorted(six.iteritems(kwargs)))
                    + (loopy.version.DATA_MODEL_VERSION,)
                    + (SUMPY_KERNEL_VERSION,)
                    + (KERNEL_VERSION,)
                    + (OPT_ENABLED,))

            try:
                result = code_cache[cache_key]
                logger.debug("%s: kernel cache hit [key=%s]" % (
                    self.name, cache_key))
                return result
            except KeyError:
                pass

        logger.info("%s: kernel cache miss" % self.name)
        if CACHING_ENABLED:
            logger.info("%s: kernel cache miss [key=%s]" % (
                self.name, cache_key))

        from pytools import MinRecursionLimit
        with MinRecursionLimit(3000):
            if OPT_ENABLED:
                knl = self.get_optimized_kernel(**kwargs)
            else:
                knl = self.get_kernel()

        if CACHING_ENABLED:
            code_cache.store_if_not_present(cache_key, knl)

        return knl


# }}} End loopy kernel cache wrapper

# {{{ scalar field expression eval


class ScalarFieldExpressionEvaluation(KernelCacheWrapper):
    """
    Evaluate a field funciton on a set of D-d points.
    Useful for imposing analytic conditions efficiently.
    """

    def __init__(self, dim, expression, variables=None, function_manglers=None):
        """
        :arg dim
        :arg expression A pymbolic expression for the function
        :arg variables A list of variables representing spacial coordinates
        """
        assert dim > 0
        self.dim = dim

        assert isinstance(expression, (ExpressionType, int, float, complex))
        self.expr = expression

        if variables is None:
            self.vars = [pmbl.var("x%d" % d) for d in range(self.dim)]
        else:
            assert isinstance(variables, list)
            for var in variables:
                assert isinstance(var, VariableType)
            self.vars = variables

        self.function_manglers = function_manglers

        self.name = "ScalarFieldExpressionEvaluation"

    def get_cache_key(self):
        return (
            type(self).__name__,
            str(self.dim) + "D",
            self.expr.__str__(),
            ",".join([x.__str__() for x in self.vars]),
        )

    def get_normalised_expr(self):
        nexpr = self.expr
        nvars = [pmbl.var("x%d" % d) for d in range(self.dim)]
        for var, nvar in zip(self.vars, nvars):
            nexpr = pmbl.substitute(nexpr, {var: nvar})
        return nexpr

    def get_variable_assignment_code(self):
        if self.dim == 1:
            return "<> x0 = target_points[0, itgt]"
        elif self.dim == 2:
            return """<> x0 = target_points[0, itgt]
                      <> x1 = target_points[1, itgt]"""
        elif self.dim == 3:
            return """<> x0 = target_points[0, itgt]
                      <> x1 = target_points[1, itgt]
                      <> x2 = target_points[2, itgt]"""
        else:
            raise NotImplementedError

    def get_kernel(self, **kwargs):

        extra_kernel_kwarg_types = ()
        if "extra_kernel_kwarg_types" in kwargs:
            extra_kernel_kwarg_types = kwargs["extra_kernel_kwarg_types"]

        eval_inames = frozenset(["itgt"])
        scalar_assignment = lp.Assignment(
            id=None,
            assignee="expr_val",
            expression=self.get_normalised_expr(),
            temp_var_type=None,
        )
        eval_insns = [
            insn.copy(within_inames=insn.within_inames | eval_inames)
            for insn in [scalar_assignment]
        ]

        loopy_knl = lp.make_kernel(  # NOQA
            "{ [itgt]: 0<=itgt<n_targets }",
            [
                """
                for itgt
                    VAR_ASSIGNMENT
                end
                """.replace(
                    "VAR_ASSIGNMENT", self.get_variable_assignment_code()
                )
            ]
            + eval_insns
            + [
                """
                for itgt
                    result[itgt] = expr_val
                end
                """
            ],
            [
                lp.ValueArg("dim, n_targets", np.int32),
                lp.GlobalArg("target_points", np.float64, "dim, n_targets"),
                ]
                + list(extra_kernel_kwarg_types)
                + ["...", ],
            name="eval_expr",
            lang_version=(2018, 2),
        )

        loopy_knl = lp.fix_parameters(loopy_knl, dim=self.dim)
        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        if self.function_manglers is not None:
            loopy_knl = lp.register_function_manglers(
                loopy_knl,
                self.function_manglers
                )

        return loopy_knl

    def get_optimized_kernel(self, ncpus=None, **kwargs):
        knl = self.get_kernel(**kwargs)
        if ncpus is None:
            import multiprocessing
            # NOTE: this detects the number of logical cores, which
            # may result in suboptimal performance.
            ncpus = multiprocessing.cpu_count()
        knl = lp.split_iname(
            knl,
            split_iname="itgt",
            inner_length=ncpus,
            inner_tag="g.0")
        return knl

    def __call__(self, queue, target_points, **kwargs):
        """
        :arg target_points
        :arg extra_kernel_kwargs
        """
        # handle target_points given as an obj_array of coords
        if (isinstance(target_points, np.ndarray)
                and target_points.dtype == np.object
                and isinstance(target_points[0], cl.array.Array)):
            target_points = cl.array.concatenate(
                    target_points).reshape([self.dim, -1])

        assert target_points.shape[0] == self.dim

        n_tgt_points = target_points[0].shape[0]
        for tgt_d in target_points:
            assert len(tgt_d) == n_tgt_points

        extra_kernel_kwargs = {}
        if "extra_kernel_kwargs" in kwargs:
            extra_kernel_kwargs = kwargs["extra_kernel_kwargs"]

        knl = self.get_cached_optimized_kernel()

        # FIXME: caching loses function mangler information
        if self.function_manglers is not None:
            knl = lp.register_function_manglers(knl, self.function_manglers)

        evt, res = knl(
            queue,
            target_points=target_points,
            n_targets=n_tgt_points,
            result=np.zeros(n_tgt_points),
            **extra_kernel_kwargs
        )

        return res["result"]

# }}} End scalar field expression eval

# {{{ import code


def import_code(code, name, add_to_sys_modules=True):
    """Dynamically generates a module.

    :arg code: can be any object containing code -- string, file object, or
    compiled code object. Returns a new module object initialized
    by dynamically importing the given code and optionally adds it
    to sys.modules under the given name.
    """
    import imp
    module = imp.new_module(name)

    if add_to_sys_modules:
        import sys
        sys.modules[name] = module

    exec(code, module.__dict__)

    return module

# }}} End import code

# {{{ box-specific maps


class BoxSpecificMap(KernelCacheWrapper):
    """
    Box-specific transform that maps between datum defined on quadrature
    nodes. Being box-specific means that the transform for each box is
    independent from the rest of the boxes.
    """
    pass

# {{{ discrete Legendre transform


class DiscreteLegendreTransform(BoxSpecificMap):
    """
    Transform from nodal values to Legendre polynomial coefficients
    for all cells (leaf boxes of a boxtree Tree object).
    It is assumed that the traversal is built over a tree where the
    sources and targets coincide.
    """

    def __init__(self, dim, degree):
        """
        :arg dim
        :arg degree Number of nodes in each axis direction.
        """
        assert dim > 0
        self.dim = dim
        assert degree > 0
        self.degree = degree

        # Template interval
        self.template_interval = [-1., 1.]
        self.template_interval_extent = 2.
        self.template_interval_center = 0.

        self.leg_tplt_x, self.leg_tplt_w = \
                np.polynomial.legendre.leggauss(degree)

        if self.dim == 1:
            self.V = np.polynomial.legendre.legvander(
                    self.leg_tplt_x, self.degree - 1)
            self.W = self.leg_tplt_w.reshape(-1)

        elif self.dim == 2:
            x, y = np.meshgrid(self.leg_tplt_x, self.leg_tplt_x)
            self.V = np.polynomial.legendre.legvander2d(
                    x.reshape(-1), y.reshape(-1),
                    [self.degree - 1] * self.dim)
            self.W = (self.leg_tplt_w[None, :]
                    * self.leg_tplt_w[:, None]).reshape(-1)

        elif self.dim == 3:
            x, y, z = np.meshgrid(self.leg_tplt_x, self.leg_tplt_x, self.leg_tplt_x)
            self.V = np.polynomial.legendre.legvander3d(
                    x.reshape(-1), y.reshape(-1), z.reshape(-1),
                    [self.degree - 1] * self.dim)
            self.W = (self.leg_tplt_w[:, None, None]
                    * self.leg_tplt_w[None, :, None]
                    * self.leg_tplt_w[None, None, :]).reshape(-1)

        else:
            raise NotImplementedError("Dimension %d is not supported" % self.dim)

        # Vandermonde matrix: each column corresponds to one basis function
        assert self.V.shape == (self.degree**self.dim, self.degree**self.dim)
        assert self.W.shape == (self.degree**self.dim,)

        # Normalizers
        self.I = np.ascontiguousarray(  # noqa: E741
                np.diag(
                    (self.V.T * self.W) @ self.V))
        assert self.I.shape == (self.degree**self.dim,)

        # Fix strides for loopy
        self.V = np.ascontiguousarray(self.V)

        # Check orthogonality
        ortho_resid = np.linalg.norm(
                self.V.T * np.matmul(self.W, self.V) - np.diag(self.I))
        if ortho_resid > 1e-13:
            logger.warn("Legendre polynomials' orthogonality residual = %f"
                    % ortho_resid)

        self.name = "DiscreteLegendreTransform"

    def get_cache_key(self):
        return (
            type(self).__name__,
            str(self.dim) + "D",
            "degree=%d" % self.degree
        )

    def get_kernel(self, **kwargs):

        loopy_knl = lp.make_kernel(  # NOQA
                [
                    "{ [ bid ] : 0 <= bid < n_boxes }",
                    "{ [ mid ] : 0 <= mid < n_box_nodes }",
                    "{ [ nid ] : 0 <= nid < n_box_nodes }"
                    ],
                [
                    """
                for bid
                    <> box_id       = boxes[bid]
                    <> box_node_beg = box_node_starts[box_id]

                    # Rescale weights based on template interval sizes.
                    # Not needed since the rscl in both the numerator and
                    # the denominator and is canceled.
                    #
                    # <> box_level    = box_levels[box_id]
                    # <> box_extent   = root_extent * (1.0 / (2**box_level))
                    # <> weight_rscl  = (box_extent / 2.0)**dim

                    for mid

                        <> mode_id = box_node_beg + mid

                        for nid
                            <> user_node_id = user_node_ids[box_node_beg + nid]
                        end

                        result[mode_id] = sum(
                                              nid,
                                              (
                                              func[user_node_id]
                                              * weight[nid]
                                              * vandermonde[nid, mid]
                                              ) * filter_multiplier[nid]
                                             ) / normalizer[mid]
                    end
                end
                """
                ],
                [
                    lp.ValueArg("n_box_nodes, n_boxes", np.int32),
                    # lp.ValueArg("root_extent", np.float64),
                    lp.GlobalArg("weight, normalizer, filter_multiplier",
                        np.float64, "n_box_nodes"),
                    lp.GlobalArg("vandermonde", np.float64,
                        "n_box_nodes, n_box_nodes"),
                    lp.GlobalArg("func", np.float64, "n_box_nodes * n_boxes"),
                    "...",
                    ],
                name="discrete_legendre_transform",
                lang_version=(2018, 2),
                )

        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        return loopy_knl

    def get_optimized_kernel(self, ncpus=None, **kwargs):
        knl = self.get_kernel(**kwargs)
        if ncpus is None:
            import multiprocessing
            ncpus = multiprocessing.cpu_count()
        knl = lp.split_iname(
            knl,
            split_iname="bid",
            inner_length=ncpus,
            inner_tag="g.0")
        return knl

    def __call__(self, queue, traversal, nodal_vals, filtering=None, **kwargs):
        """
        :arg traversal
        :arg nodal_vals CL array of nodal values.
        :arg filtering Box-wide filter given by an CL array or None.
        """

        if filtering is None:
            filter_multiplier = 1 + cl.array.zeros(queue,
                    self.degree**self.dim, np.float64)
        elif isinstance(filtering, cl.array.Array):
            assert filtering.shape == (self.degree**self.dim,)
            filter_multiplier = filtering
        else:
            raise RuntimeError("Invalid filtering argument: %s"
                    % str(filtering))

        knl = self.get_cached_optimized_kernel()

        evt, res = knl(
            queue,
            boxes=traversal.target_boxes,
            box_node_starts=traversal.tree.box_target_starts,
            user_node_ids=traversal.tree.user_source_ids,
            # box_levels=traversal.tree.box_levels,
            # root_extent=traversal.tree.root_extent,
            func=nodal_vals,
            weight=cl.array.to_device(queue, self.W),
            vandermonde=cl.array.to_device(queue, self.V),
            normalizer=cl.array.to_device(queue, self.I),
            n_box_nodes=self.degree**self.dim,
            n_boxes=traversal.target_boxes.shape[0],
            filter_multiplier=filter_multiplier,
            result=cl.array.zeros_like(nodal_vals),
        )

        return res["result"]

# }}} End discrete Legendre transform

# }}} End box-specific maps

# {{{ box-specific reductions


class BoxSpecificReduction(KernelCacheWrapper):
    """
    Box-specific reduction that maps for each box a data vector defined
    on the quadrature nodes to a scalar.
    Being box-specific means that the reductions for each box is
    independent from the rest of the boxes.
    """
    pass

# {{{ sum


class BoxSum(BoxSpecificReduction):
    """
    Adds up nodal values within each box.
    """

    def __init__(self, dim, degree):
        """
        :arg dim
        :arg degree Number of nodes in each axis direction.
        """
        assert dim > 0
        self.dim = dim
        assert degree > 0
        self.degree = degree

        self.name = "BoxSum"

    def get_cache_key(self):
        return (
            type(self).__name__,
            str(self.dim) + "D",
            "degree=%d" % self.degree
        )

    def get_kernel(self, **kwargs):

        loopy_knl = lp.make_kernel(  # NOQA
                [
                    "{ [ bid ] : 0 <= bid < n_boxes }",
                    "{ [ nid ] : 0 <= nid < n_box_nodes }"
                    ],
                [
                    """
                for bid
                    <> box_id       = boxes[bid]
                    <> box_node_beg = box_node_starts[box_id]

                    result[bid] = sum(nid,
                                      func[box_node_beg + nid]
                                      * filter_multiplier[nid])
                end
                """
                ],
                [
                    lp.ValueArg("n_box_nodes, n_boxes", np.int32),
                    lp.GlobalArg("filter_multiplier", np.float64, "n_box_nodes"),
                    lp.GlobalArg("func", np.float64, "n_box_nodes * n_boxes"),
                    "...",
                    ],
                name="box_filtered_sum",
                lang_version=(2018, 2),
                )

        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        return loopy_knl

    def get_optimized_kernel(self, ncpus=None, **kwargs):
        knl = self.get_kernel(**kwargs)
        if ncpus is None:
            import multiprocessing
            ncpus = multiprocessing.cpu_count()
        knl = lp.split_iname(
            knl,
            split_iname="bid",
            inner_length=ncpus,
            inner_tag="g.0")
        return knl

    def __call__(self, queue, traversal, nodal_vals, filtering=None, **kwargs):
        """
        :arg traversal
        :arg nodal_vals CL array of nodal values.
        :arg filtering Box-wide filter given by an CL array or None.

        .. warning::
           The output of this kernel is ordered in :mod:`boxtree`'s box ids.
           It may not be the same as the order implied by the input (e.g.
           box mesh generated by dealii).
        """

        if filtering is None:
            filter_multiplier = 1 + cl.array.zeros(queue,
                    self.degree**self.dim, np.float64)
        elif isinstance(filtering, cl.array.Array):
            assert filtering.shape == (self.degree**self.dim,)
            filter_multiplier = filtering
        else:
            raise RuntimeError("Invalid filtering argument: %s"
                    % str(filtering))

        knl = self.get_cached_optimized_kernel()
        n_boxes = traversal.target_boxes.shape[0]

        evt, res = knl(
            queue,
            boxes=traversal.target_boxes,
            box_node_starts=traversal.tree.box_target_starts,
            func=nodal_vals,
            n_box_nodes=self.degree**self.dim,
            n_boxes=n_boxes,
            filter_multiplier=filter_multiplier,
            result=cl.array.zeros(queue, n_boxes, nodal_vals.dtype),
        )

        return res["result"]

# }}} End sum

# }}} End box-specific reductions

# {{{ filters for box-specific operators


def generate_leading_order_filtering(dim, n_dofs):
    """Returns a filtering vector that is an indicator function of the node
    that corresponds to the leading order modal values in the Fourier space.
    """

    mask1d = np.zeros(n_dofs)
    mask1d[-1] = 1

    if dim == 1:
        return mask1d

    elif dim == 2:
        return (
            mask1d[:, None] + mask1d[None, :]
            - mask1d[:, None] * mask1d[None, :]
        ).reshape(-1)

    elif dim == 3:
        return (
            mask1d[:, None, None] + mask1d[None, :, None] + mask1d[None, None, :]
            - mask1d[:, None, None] * mask1d[None, :, None]
            - mask1d[:, None, None] * mask1d[None, None, :]
            - mask1d[None, :, None] * mask1d[None, None, :]
            + mask1d[:, None, None] * mask1d[None, :, None] * mask1d[None, None, :]
        ).reshape(-1)

    else:
        raise NotImplementedError("Dimension %d not supported" % dim)

# }}} End filters for box-specific operators



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# {{{ find install- or run-time git revision

import os
if os.environ.get("AKPYTHON_EXEC_FROM_WITHIN_WITHIN_SETUP_PY") is not None:
    # We're just being exec'd by setup.py. We can't import anything.
    _git_rev = None

else:
    import volumential._git_rev as _git_rev_mod
    _git_rev = _git_rev_mod.GIT_REVISION

    # If we're running from a dev tree, the last install (and hence the most
    # recent update of the above git rev) could have taken place very long ago.
    from pytools import find_module_git_revision
    _runtime_git_rev = find_module_git_revision(__file__, n_levels_up=1)
    if _runtime_git_rev is not None:
        _git_rev = _runtime_git_rev

# }}}

VERSION = (2017, 1)
VERSION_STATUS = "alpha"
VERSION_TEXT = ".".join(str(i) for i in VERSION) + VERSION_STATUS

KERNEL_VERSION = (VERSION, _git_rev, 0)

LOOPY_LANG_VERSION = (2018, 2)

# vim: ft=pyopencl



from __future__ import absolute_import, division, print_function

__doc__ = """
.. autofunction:: generate_list1_gallery
"""

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np


# {{{ TreeBox class
class TreeBox:
    """
    A minimalistic tree class.
    Courtesy of: Andreas Klockner
    """

    def __init__(self, center, radius, child_nlevels):
        self.center = center
        self.radius = radius

        self.children = []

        if child_nlevels:
            child_radius = radius // 2
            assert child_radius

            dimensions = len(center)

            for i in range(2 ** dimensions):
                child_center = center.copy()
                for idim in range(dimensions):
                    # 1 if that dimension bit is set, 0 if not
                    dim_indicator = int(bool(i & 1 << idim))
                    child_center[idim] += (2 * dim_indicator - 1) * child_radius

                self.children.append(
                    TreeBox(child_center, child_radius, child_nlevels - 1)
                )

    def draw(self):
        lx, ly = self.center - self.radius * 0.95
        hx, hy = self.center + self.radius * 0.95

        import matplotlib.pyplot as plt

        plt.plot([lx, lx, hx, hx, lx], [ly, hy, hy, ly, ly])


# }}} End TreeBox class


def build_tree(dimensions):
    """
    Courtesy of: Andreas Klockner
    """
    # four levels deep
    # -> centers on a 2**4 x 2**4 grid

    nlevels = 4
    root_radius = 2 ** (nlevels - 1)
    root = TreeBox(
        center=np.array(dimensions * [root_radius], np.int),
        radius=root_radius,
        child_nlevels=nlevels - 1,
    )

    return root


def generate_boxes_on_level(box, ilevel):
    """
    Courtesy of: Andreas Klockner
    """
    if ilevel:
        for child in box.children:
            for result in generate_boxes_on_level(child, ilevel - 1):
                yield result
    else:
        yield box


def generate_boxes(box):
    """
    Courtesy of: Andreas Klockner
    """
    yield box

    for child in box.children:
        for result in generate_boxes(child):
            yield result


def linf_dist(box1, box2):
    """
    Courtesy of: Andreas Klockner
    """
    return np.max(np.abs(box1.center - box2.center) - (box1.radius + box2.radius))


def generate_interactions(dimensions):
    """
    Courtesy of: Andreas Klockner
    """
    root = build_tree(dimensions)
    root_radius = root.radius

    min_cutoff = root_radius >> 2
    max_cutoff = 2 * root_radius - min_cutoff

    target_boxes = [
        box
        for box in generate_boxes_on_level(root, 2)
        if np.min(box.center) > min_cutoff and np.max(box.center) < max_cutoff
    ]

    near_neighbor_interactions = [
        (tbox, sbox)
        for tbox in target_boxes
        for sbox in generate_boxes(root)
        if linf_dist(tbox, sbox) == 0
    ]

    if 0:
        import matplotlib.pyplot as plt

        for tbox, sbox in near_neighbor_interactions:
            plt.figure()
            plt.gca().set_aspect("equal")
            tbox.draw()
            sbox.draw()
    return near_neighbor_interactions


def postprocess_interactions(near_neighbor_interactions):
    unique_interaction_vectors = set()

    for (tbox, sbox) in near_neighbor_interactions:
        unique_interaction_vectors.add(tuple(tbox.center - sbox.center))

    # Add interactions within the same box
    tb0, wb0 = near_neighbor_interactions[0]
    unique_interaction_vectors.add(tuple(tb0.center - tb0.center))

    list1_interactions = sorted(list(unique_interaction_vectors))

    return list1_interactions


def generate_list1_gallery(dim):
    """Generate a *list1* that servers as the gallery for all possible *list1*
    interactions with given dimension and order.
    """
    # contains each sourcebox.center-to-targetbox.center vector
    # source box is 4x4 to make all involved lengths to be integers
    vec_list = postprocess_interactions(generate_interactions(dim))

    distinct_numbers = set()
    for vec in vec_list:
        for l in vec:
            distinct_numbers.add(l)

    # contains a lookup table for case indices
    base = len(range(min(distinct_numbers), max(distinct_numbers) + 1))
    case_indices = -np.ones(base ** dim, dtype=int)
    shift = -min(distinct_numbers)

    def case_encode(case_vec):
        table_id = 0
        for l in case_vec:
            table_id = table_id * base + (l + shift)
        return int(table_id)

    case_id = 0
    for vec in vec_list:
        case_indices[case_encode(vec)] = case_id
        case_id += 1

    assert len(vec_list) == case_id

    return (vec_list, case_encode, case_indices)


# vim: ft=pyopencl:fdm=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import volumential.nearfield_potential_table as npt
import numpy as np
from numpy.polynomial.chebyshev import chebval, chebval2d, chebval3d


def test_const_order_1():
    table = npt.NearFieldInteractionTable(quad_order=1, build_method='Transform')
    table.build_table()
    for ary in table.data:
        assert np.allclose(ary, 1)


def test_const_order_2(longrun):
    table = npt.NearFieldInteractionTable(quad_order=2, build_method='Transform')
    table.build_table()
    for ary in table.data:
        assert np.allclose(ary, 0.25)


def interp_modes(q_order):
    table = npt.NearFieldInteractionTable(
            quad_order=q_order, build_method='Transform')

    modes = [table.get_mode(i) for i in range(table.n_q_points)]

    def interpolate_function(x, y):
        return sum([mode(x, y) for mode in modes])

    h = 0.2
    xx = yy = np.arange(-1.0, 1.0, h)
    xi, yi = np.meshgrid(xx, yy)
    xi = xi.flatten()
    yi = yi.flatten()

    val = np.zeros(xi.shape)
    print(xi)
    print(yi)

    for i in range(len(xi)):
        val[i] = interpolate_function(xi[i], yi[i])

    return val


def test_modes():
    for q in [1, 2, 3, 4, 5, 6, 7, 8]:
        val = interp_modes(q)
        assert np.allclose(val, 1)


def cheb_eval(dim, coefs, coords):
    if dim == 1:
        return chebval(coords[0], coefs)
    elif dim == 2:
        return chebval2d(coords[0], coords[1], coefs)
    elif dim == 3:
        return chebval3d(coords[0], coords[1], coords[2], coefs)
    else:
        raise NotImplementedError('dimension %d not supported' % dim)


def drive_test_modes_cheb_coeffs(dim, q, cheb_order):
    if not cheb_order >= q:
        raise RuntimeError('Insufficient cheb_order to fully resolve the modes')

    sample_mode = np.random.randint(q**dim)
    table = npt.NearFieldInteractionTable(quad_order=q, dim=dim)
    ccoefs = table.get_mode_cheb_coeffs(sample_mode, cheb_order)
    shape = (cheb_order, ) * dim
    ccoefs = ccoefs.reshape(*shape)

    # Evaluate the mode at the interpolation nodes via
    # inverse Chebyshev transform.
    #
    # NOTE: table.q_points are over [0, 1]^dim,
    # while cheb_eval assumes points are over [-1, 1]^dim
    targets = np.array([
        [qpt[i] for qpt in table.q_points]
        for i in range(dim)]) * 2 - 1

    mode_vals = cheb_eval(dim, ccoefs, targets)
    mode_vals[np.abs(mode_vals) < 8 * np.finfo(mode_vals.dtype).eps] = 0

    mode_interp_vals = np.zeros(q**dim)
    mode_interp_vals[sample_mode] = 1

    assert np.allclose(mode_vals, mode_interp_vals)


def test_modes_cheb_coeffs():
    drive_test_modes_cheb_coeffs(1, 5, 5)
    drive_test_modes_cheb_coeffs(1, 5, 10)
    drive_test_modes_cheb_coeffs(1, 10, 10)
    drive_test_modes_cheb_coeffs(1, 15, 15)

    drive_test_modes_cheb_coeffs(2, 5, 5)
    drive_test_modes_cheb_coeffs(2, 5, 10)
    drive_test_modes_cheb_coeffs(2, 10, 10)
    drive_test_modes_cheb_coeffs(2, 15, 15)

    drive_test_modes_cheb_coeffs(3, 5, 5)
    drive_test_modes_cheb_coeffs(3, 5, 10)
    drive_test_modes_cheb_coeffs(3, 10, 10)


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from py.test.cmdline import main

        main([__file__])

# vim: foldmethod=marker:filetype=pyopencl



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import loopy as lp
from volumential.tools import KernelCacheWrapper

import logging

logger = logging.getLogger(__name__)

__doc__ = """
.. autoclass:: DrosteBase
   :members:
.. autoclass:: DrosteFull
   :members:
.. autoclass:: DrosteReduced
   :members:
.. autoclass:: InverseDrosteReduced
   :members:
"""

if 0:
    logging.basicConfig(level=logging.INFO)

# {{{ Droste base class


class DrosteBase(KernelCacheWrapper):
    """
    Base class for Droste methods.
    It uses sumpy tools to cache the loopy kernel.

    .. attribute:: integral_knl

       The integral kernel of sumpy.kernel type.

    .. attribute:: interaction_case_vecs

       The relative positions of the target box for each case.

    .. attribute:: interaction_case_scls

       The relative sizes of the target box for each case.
    """

    def __init__(self, integral_knl, quad_order, case_vecs, n_brick_quad_points=15):
        from sumpy.kernel import Kernel

        if integral_knl is not None:
            assert isinstance(integral_knl, Kernel)
            self.integral_knl = integral_knl
            self.dim = integral_knl.dim
        else:
            self.integral_knl = None
            self.dim = None

        if case_vecs is not None:
            self.ncases = len(case_vecs)
            for cvec in case_vecs:
                assert len(cvec) == self.dim
            self.interaction_case_vecs = np.array(
                [[v[d] for v in case_vecs] for d in range(self.dim)]
            )
        else:
            self.ncases = 0
            case_vecs = list()
            self.interaction_case_vecs = np.array([])

        self.interaction_case_scls = np.array(
            [
                1
                if int(max(abs(np.array(vec)))) == 0
                else max([abs(l) - 0.5 for l in np.array(vec) / 4]) * 2
                for vec in case_vecs
            ]
        )

        self.nfunctions = quad_order
        self.ntgt_points = quad_order
        self.nquad_points = n_brick_quad_points

        if self.dim is not None:
            self.n_q_points = quad_order ** self.dim
        else:
            self.n_q_points = None

        self.name = "DrosteBase"

    def make_basis_vars(self):
        d = self.dim
        self.basis_vars = ["f%d" % i for i in range(d)]
        return self.basis_vars

    def make_tgt_vars(self):
        d = self.dim
        self.tgt_vars = ["t%d" % i for i in range(d)]
        return self.tgt_vars

    def make_quad_vars(self):
        d = self.dim
        self.quad_vars = ["q%d" % i for i in range(d)]
        return self.quad_vars

    def make_basis_eval_vars(self):
        d = self.dim
        self.basis_eval_vars = ["p%d" % i for i in range(d)]
        return self.basis_eval_vars

    def make_pwaffs(self):
        import islpy as isl

        self.pwaffs = isl.make_zero_and_vars(
            self.tgt_vars
            + self.basis_vars
            + self.basis_eval_vars
            + self.quad_vars
            + ["iside", "iaxis", "ibrick_side", "ibrick_axis", "ilevel", "icase"],
            ["nlevels"],
        )
        return self.pwaffs

    def make_brick_domain(self, variables, n, lbound=0, ubound=0):
        pwaffs = self.pwaffs
        if isinstance(n, str):
            ubound_pwaff = pwaffs[n] + ubound
        else:
            ubound_pwaff = pwaffs[0] + n

        from functools import reduce
        import operator

        return reduce(
            operator.and_,
            [(pwaffs[0] + lbound).le_set(pwaffs[var]) for var in variables]
            + [pwaffs[var].lt_set(ubound_pwaff) for var in variables],
        )

    def get_sumpy_kernel_insns(self):
        # get sumpy kernel insns
        from sumpy.symbolic import make_sym_vector

        dvec = make_sym_vector("dist", self.dim)
        from sumpy.assignment_collection import SymbolicAssignmentCollection

        sac = SymbolicAssignmentCollection()
        result_name = sac.assign_unique(
            "knl_val",
            self.integral_knl.postprocess_at_target(
                self.integral_knl.postprocess_at_source(
                    self.integral_knl.get_expression(dvec), dvec
                ),
                dvec,
            ),
        )
        sac.run_global_cse()
        import six
        from sumpy.codegen import to_loopy_insns

        loopy_insns = to_loopy_insns(
            six.iteritems(sac.assignments),
            vector_names=set(["dist"]),
            pymbolic_expr_maps=[self.integral_knl.get_code_transformer()],
            retain_names=[result_name],
            complex_dtype=np.complex128,
        )
        # for i in loopy_insns:
        #     print(i)
        return loopy_insns

    def get_sumpy_kernel_eval_insns(self):
        # actual integral kernel evaluation (within proper inames)
        knl_loopy_insns = self.get_sumpy_kernel_insns()
        quad_inames = frozenset(
            ["icase"]
            + self.tgt_vars
            + self.basis_vars
            + ["ilevel", "ibrick_axis", "ibrick_side"]
            + self.quad_vars
        )
        quad_kernel_insns = [
            insn.copy(within_inames=insn.within_inames | quad_inames)
            for insn in knl_loopy_insns
        ]

        from sumpy.symbolic import SympyToPymbolicMapper

        sympy_conv = SympyToPymbolicMapper()
        scaling_assignment = lp.Assignment(
            id=None,
            assignee="knl_scaling",
            expression=sympy_conv(self.integral_knl.get_global_scaling_const()),
            temp_var_type=lp.Optional(),
        )

        return quad_kernel_insns + [scaling_assignment]

    def codegen_basis_eval(self, iaxis):
        """Generate instructions to evaluate Chebyshev polynomial basis.
        (Chebyshev polynomials of the first kind T_n).
        """
        code = """
            <> T0_IAXIS = 1
            <> T1_IAXIS = template_mapped_point_tmp[IAXIS] {dep=mpoint}
            <> Tprev_IAXIS = T0_IAXIS {id=t0_IAXIS}
            <> Tcur_IAXIS = T1_IAXIS {id=t1_IAXIS,dep=t0_IAXIS}

            for pIAXIS
                <> Tnext_IAXIS = (2 * template_mapped_point_tmp[IAXIS] * Tcur_IAXIS
                                - Tprev_IAXIS) {id=tnextIAXIS,dep=t1_IAXIS}
                Tprev_IAXIS = Tcur_IAXIS {id=tprev_updateIAXIS,dep=tnextIAXIS}
                Tcur_IAXIS = Tnext_IAXIS {id=tcur_updateIAXIS,dep=tprev_updateIAXIS}
            end

            <> basis_evalIAXIS = (
                T0_IAXIS * if(fIAXIS == 0, 1, 0)
                + T1_IAXIS * if(fIAXIS == 1, 1, 0)
                + simul_reduce(sum, pIAXIS,
                    if(fIAXIS >= 2 and fIAXIS == pIAXIS + 2, Tcur_IAXIS, 0))
                ) {id=basisIAXIS,dep=tcur_updateIAXIS}
            """.replace(
            "IAXIS", str(iaxis)
        )
        return code

    def make_dim_independent(self, knlstring):

        # replace REV_* first
        resknl = knlstring.replace("REV_BASIS_VARS", ",".join(self.basis_vars[::-1]))
        resknl = resknl.replace("REV_TGT_VARS", ",".join(self.tgt_vars[::-1]))

        resknl = resknl.replace("BASIS_VARS", ",".join(self.basis_vars))
        resknl = resknl.replace("TGT_VARS", ",".join(self.tgt_vars))
        resknl = resknl.replace("QUAD_VARS", ",".join(self.quad_vars))

        resknl = resknl.replace(
                "POSTPROCESS_KNL_VAL",
                "<> knl_val_post = knl_val {id=pp_kval}"
                )

        resknl = resknl.replace(
            "PROD_QUAD_WEIGHT",
            " * ".join(
                [
                    "quadrature_weights[QID]".replace("QID", qvar)
                    for qvar in self.quad_vars
                ]
            ),
        )

        if self.dim == 1:
            resknl = resknl.replace("TPLTGT_ASSIGNMENT", """target_nodes[t0]""")
            resknl = resknl.replace("QUAD_PT_ASSIGNMENT", """quadrature_nodes[q0]""")
            resknl = resknl.replace("DENSITY_VAL_ASSIGNMENT", """basis_eval0""")
            resknl = resknl.replace(
                "PREPARE_BASIS_VALS",
                "\n".join([self.codegen_basis_eval(i) for i in range(self.dim)])
                + """
                ... nop {id=basis_evals,dep=basis0}
                """,
            )

        elif self.dim == 2:
            resknl = resknl.replace(
                "TPLTGT_ASSIGNMENT",
                """if(iaxis == 0, target_nodes[t0], target_nodes[t1])""",
            )
            resknl = resknl.replace(
                "QUAD_PT_ASSIGNMENT",
                """if(iaxis == 0, quadrature_nodes[q0], quadrature_nodes[q1])""",
            )
            resknl = resknl.replace(
                "DENSITY_VAL_ASSIGNMENT", """basis_eval0 * basis_eval1"""
            )
            resknl = resknl.replace(
                "PREPARE_BASIS_VALS",
                "\n".join([self.codegen_basis_eval(i) for i in range(self.dim)])
                + """
                ... nop {id=basis_evals,dep=basis0:basis1}
                """,
            )

        elif self.dim == 3:
            resknl = resknl.replace(
                "TPLTGT_ASSIGNMENT",
                """if(iaxis == 0, target_nodes[t0], if(
                          iaxis == 1, target_nodes[t1], target_nodes[t2]))""",
            )
            resknl = resknl.replace(
                "QUAD_PT_ASSIGNMENT",
                """if(iaxis == 0, quadrature_nodes[q0], if(
                  iaxis == 1, quadrature_nodes[q1], quadrature_nodes[q2]))""",
            )
            resknl = resknl.replace(
                "DENSITY_VAL_ASSIGNMENT",
                """basis_eval0 * basis_eval1 * basis_eval2"""
            )
            resknl = resknl.replace(
                "PREPARE_BASIS_VALS",
                "\n".join([self.codegen_basis_eval(i) for i in range(self.dim)])
                + """
                ... nop {id=basis_evals,dep=basis0:basis1:basis2}
                """,
            )

        else:
            raise NotImplementedError

        return resknl

    def make_result_array(self, **kwargs):
        """Allocate memory space for results.
        """
        # by default uses double type returns
        if "result_dtype" in kwargs:
            result_dtype = kwargs["result_dtype"]
        else:
            result_dtype = np.float64

        # allocate return arrays
        if self.dim == 1:
            result_array = (
                    np.zeros(
                        (self.nfunctions, self.ntgt_points, self.ncases),
                        result_dtype
                        )
                    + np.nan)
        elif self.dim == 2:
            result_array = (
                    np.zeros(
                        (
                            self.nfunctions,
                            self.nfunctions,
                            self.ntgt_points,
                            self.ntgt_points,
                            self.ncases,
                            ),
                        result_dtype
                        )
                    + np.nan
                    )
        elif self.dim == 3:
            result_array = (
                    np.zeros(
                        (
                            self.nfunctions,
                            self.nfunctions,
                            self.nfunctions,
                            self.ntgt_points,
                            self.ntgt_points,
                            self.ntgt_points,
                            self.ncases,
                            ),
                        result_dtype
                        )
                    + np.nan
                    )
        else:
            raise NotImplementedError
        return result_array

    def get_kernel_code(self):
        return [
            self.make_dim_independent(
                """  # noqa
        for iaxis
            <> root_center[iaxis] = 0.5 * (
                    root_brick[iaxis, 1] + root_brick[iaxis, 0]) {dup=iaxis}
            <> root_extent[iaxis] = (root_brick[iaxis, 1]
                    - root_brick[iaxis, 0]) {dup=iaxis}
        end

        for ilevel, BASIS_VARS, TGT_VARS, icase

            # Targets outside projected onto the boundary
            for iaxis
                <> template_target[iaxis] = TPLTGT_ASSIGNMENT \
                        {id=tplt_tgt_pre,dup=iaxis}
            end

            # True targets are used for kernel evaluation
            for iaxis
                <> true_target[iaxis] = ( root_center[iaxis]
                         + interaction_case_vecs[iaxis, icase]
                            * 0.25 * root_extent[iaxis]
                         + interaction_case_scls[icase]
                            * (template_target[iaxis] - 0.5)
                            * root_extent[iaxis]
                         ) {id=true_targets,dup=iaxis,dep=tplt_tgt_pre}

                # Re-map the mapped points to [-1,1]^dim for Chebyshev evals
                <> template_true_target[iaxis] = 0.0 + (
                        true_target[iaxis] - root_center[iaxis]
                        ) / root_extent[iaxis] * 2 {id=template_true_targets,dup=iaxis,dep=true_targets}
            end

            # Projected targets are used for brick construction
            for iaxis
                <> target[iaxis] = if(
                    true_target[iaxis] > root_brick[iaxis, 1],
                    root_brick[iaxis, 1],
                    if(
                        true_target[iaxis] < root_brick[iaxis, 0],
                        root_brick[iaxis, 0],
                        true_target[iaxis])) {dup=iaxis,dep=true_targets}
            end

            for iaxis
                template_target[iaxis] = (0.5
                    + interaction_case_vecs[iaxis, icase] * 0.25
                    + interaction_case_scls[icase] * (template_target[iaxis] - 0.5)
                    ) {id=tplt_tgt3,dup=iaxis,dep=tplt_tgt_pre:true_targets}
            end

            for iaxis
                template_target[iaxis] = if(
                    template_target[iaxis] > 1,
                    1,
                    if(
                        template_target[iaxis] < 0,
                        0,
                        template_target[iaxis])
                    ) {id=tplt_tgt4,dup=iaxis,dep=tplt_tgt3}
            end

            ... nop {id=tplt_tgt,dep=tplt_tgt_pre:tplt_tgt4:true_targets}

            # Debug output
            #for iaxis
            #    projected_template_targets[TGT_VARS, icase, iaxis] = \
            #            template_target[iaxis] {dup=iaxis,dep=tplt_tgt}
            #end

            # Debug output
            #for iaxis
            #    target_points[ilevel, BASIS_VARS, TGT_VARS, icase, iaxis
            #        ] = true_target[iaxis] {dup=iaxis,dep=true_targets}
            #end

            for iaxis, iside
                <> outer_brick[iaxis, iside] = (
                        (alpha**ilevel)*root_brick[iaxis,iside]
                        + (1-alpha**ilevel) * target[iaxis])  {dup=iaxis:iside}
                <> inner_brick[iaxis, iside] = (
                        alpha*outer_brick[iaxis,iside]
                        + (1-alpha) * target[iaxis])  {dup=iaxis:iside}
            end

            for iaxis
                <> ob_ext[iaxis] = (outer_brick[iaxis, 1]
                    - outer_brick[iaxis, 0])  {dup=iaxis}
            end

            for ibrick_axis, ibrick_side, QUAD_VARS
                for iaxis
                    <> point[iaxis] = QUAD_PT_ASSIGNMENT {dup=iaxis}
                end

                <> deform = point[ibrick_axis]*(1-alpha)

                for iaxis
                    if iaxis == ibrick_axis
                        <> mapped_point_tmp[iaxis] = (
                            point[ibrick_axis]*(
                                inner_brick[ibrick_axis, ibrick_side]
                                - outer_brick[ibrick_axis, ibrick_side])
                                + outer_brick[ibrick_axis, ibrick_side]) \
                            {id=mpoint1,nosync=mpoint2}
                    else
                        <> pre_scale = (
                            point[iaxis]
                            + deform*(template_target[iaxis]-point[iaxis])
                            ) {dep=tplt_tgt}

                        mapped_point_tmp[iaxis] = \
                            ob_ext[iaxis] * pre_scale + outer_brick[iaxis, 0] \
                            {id=mpoint2,nosync=mpoint1}
                    end

                    ... nop {id=mpoint,dep=mpoint1:mpoint2}

                    # Re-map the mapped points to [-1,1]^dim for Chebyshev evals
                    <> template_mapped_point_tmp[iaxis] = 0.0 + (
                            mapped_point_tmp[iaxis] - root_center[iaxis]
                            ) / root_extent[iaxis] * 2 {dep=mpoint}

                    # Debug output
                    #mapped_points[ilevel, ibrick_axis,
                    #          ibrick_side, TGT_VARS, icase, iaxis, QUAD_VARS] = \
                    #    mapped_point_tmp[iaxis]  {dep=mpoint}
                    #mapped_target[TGT_VARS, icase, iaxis] = target[iaxis]
                end

                for iaxis
                    if iaxis == ibrick_axis
                        <> jac_part = (
                            inner_brick[ibrick_axis, ibrick_side]
                            - outer_brick[ibrick_axis, ibrick_side]
                            )    {id=jpart1}
                    else
                        jac_part = ob_ext[iaxis] * (1-deform)  {id=jpart2}
                    end
                end

                <> jacobian = abs(product(iaxis,
                                          jac_part)) {id=jac,dep=jpart1:jpart2}

                <> dist[iaxis] = (true_target[iaxis]
                                - mapped_point_tmp[iaxis]) {dep=mpoint:true_targets}

                # optional postprocessing of kernel values
                POSTPROCESS_KNL_VAL

                # in our case 0 * inf = 0
                <> knl_val_finished = if(abs(knl_val_post) > 1e16,
                    0, knl_val_post) {id=finish_kval,dep=pp_kval}

            end

        end

        for BASIS_VARS, TGT_VARS, icase
            for ilevel
                for ibrick_axis, ibrick_side, QUAD_VARS

                    PREPARE_BASIS_VALS

                    <> density_val = DENSITY_VAL_ASSIGNMENT \
                            {id=density,dep=basis_evals}

                end
            end
        end

        for BASIS_VARS, TGT_VARS, icase
            # the first index is contiguous
            result[REV_BASIS_VARS, REV_TGT_VARS, icase] = (sum(
                    (ilevel, ibrick_axis, ibrick_side, QUAD_VARS),
                    PROD_QUAD_WEIGHT
                    * jacobian
                    * knl_val_finished
                    * density_val
                    )
                    *  knl_scaling
                ) {id=result,dep=jac:mpoint:density:finish_kval}
        end
        """
            )
        ]

    def get_target_points(self, queue=None):
        import volumential.meshgen as mg

        q_points, _, _ = mg.make_uniform_cubic_grid(
            degree=self.ntgt_points, level=1, dim=self.dim,
            queue=queue)

        # map to [0,1]^d
        mapped_q_points = np.array(
                [
                    0.5 * (qp + np.ones(self.dim))
                    for qp in q_points
                    ]
                )
        # sort in dictionary order, preserve only the leading
        # digits to prevent floating point errors from polluting
        # the ordering.
        q_points_ordering = sorted(
            range(len(mapped_q_points)),
            key=lambda i: list(np.floor(mapped_q_points[i] * 10000)),
        )
        return mapped_q_points[q_points_ordering]

    def postprocess_cheb_table(self, cheb_table, cheb_coefs):
        nfp_table = np.zeros([self.n_q_points, ] + list(cheb_table.shape[self.dim:]))

        # transform to interpolatory basis functions
        # NOTE: the reversed order of indices, e.g.,
        #       mccoefs[f0, f1, f2], and cheb_table[f2, f1, f0]
        concat_axes = [iaxis for iaxis in range(self.dim)]
        for mid in range(self.n_q_points):
            mccoefs = cheb_coefs[mid]
            nfp_table[mid] = np.tensordot(
                mccoefs.reshape([self.nfunctions for iaxis in range(self.dim)]),
                cheb_table,
                axes=(concat_axes, concat_axes[::-1]),
            )

        # transform to self.data format (icase, source_id, target_id)
        # NOTE: the directions of target_id are reversed since cheb_table
        #       is indexed by, e.g. cheb_table[t2, t1, t0]
        transpose_axes = (self.dim + 1, 0) + tuple(
            self.dim - i for i in range(self.dim)
        )

        return nfp_table.transpose(transpose_axes).reshape(-1, order="C")


# }}} End Droste base class

# {{{ full Droste method


class DrosteFull(DrosteBase):
    """
    Build the full table directly.
    """

    def __init__(self, integral_knl, quad_order, case_vecs, n_brick_quad_points=50):
        super(DrosteFull, self).__init__(
                integral_knl, quad_order, case_vecs, n_brick_quad_points)
        self.name = "DrosteFull"

    def make_loop_domain(self):
        tgt_vars = self.make_tgt_vars()
        quad_vars = self.make_quad_vars()
        basis_vars = self.make_basis_vars()
        basis_eval_vars = self.make_basis_eval_vars()
        pwaffs = self.make_pwaffs()  # noqa: F841
        self.loop_domain = (
            self.make_brick_domain(tgt_vars, self.ntgt_points)
            & self.make_brick_domain(quad_vars, self.nquad_points)
            & self.make_brick_domain(basis_vars, self.nfunctions)
            & self.make_brick_domain(basis_eval_vars, self.nfunctions)
            & self.make_brick_domain(["iside"], 2)
            & self.make_brick_domain(["iaxis"], self.dim)
            & self.make_brick_domain(["ibrick_side"], 2)
            & self.make_brick_domain(["ibrick_axis"], self.dim)
            & self.make_brick_domain(["ilevel"], "nlevels")
            & self.make_brick_domain(["icase"], self.ncases)
        )
        return self.loop_domain

    def get_kernel(self, **kwargs):

        domain = self.make_loop_domain()

        extra_kernel_kwarg_types = ()
        if "extra_kernel_kwarg_types" in kwargs:
            extra_kernel_kwarg_types = kwargs["extra_kernel_kwarg_types"]

        loopy_knl = lp.make_kernel(  # NOQA
            [domain],
            self.get_kernel_code() + self.get_sumpy_kernel_eval_insns(),
            [
                lp.ValueArg("alpha", np.float64),
                lp.ValueArg("n_cases, nfunctions, quad_order, dim", np.int32),
                lp.GlobalArg("interaction_case_vecs", np.float64, "dim, n_cases"),
                lp.GlobalArg("interaction_case_scls", np.float64, "n_cases"),
                lp.GlobalArg(
                    "result", None,
                    ", ".join(
                        ["nfunctions" for d in range(self.dim)]
                        + ["quad_order" for d in range(self.dim)]
                    )
                    + ", n_cases",
                ), ]
                + list(extra_kernel_kwarg_types)
                + ["...", ],
            name="brick_map",
            lang_version=(2018, 2),
        )

        loopy_knl = lp.fix_parameters(loopy_knl, d=self.dim)
        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        try:
            loopy_knl = self.integral_knl.prepare_loopy_kernel(loopy_knl)
        except Exception:
            pass

        return loopy_knl

    def get_optimized_kernel(self, ncpus=None, **kwargs):
        if ncpus is None:
            import multiprocessing
            # NOTE: this detects the number of logical cores, which
            # may result in suboptimal performance.
            ncpus = multiprocessing.cpu_count()
        knl = self.get_kernel(**kwargs)
        knl = lp.split_iname(knl, "icase", ncpus, inner_tag="g.0")
        return knl

    def call_loopy_kernel(self, queue, **kwargs):
        """
        :arg source_box_extent
        :arg alpha
        :arg nlevels
        :arg extra_kernel_kwargs
        """

        if "source_box_extent" in kwargs:
            assert kwargs["source_box_extent"] > 0
            source_box_extent = kwargs["source_box_extent"]
        else:
            source_box_extent = 1

        extra_kernel_kwargs = {}
        if "extra_kernel_kwargs" in kwargs:
            extra_kernel_kwargs = kwargs["extra_kernel_kwargs"]

        if "alpha" in kwargs:
            alpha = kwargs["alpha"]
            assert alpha >= 0 and alpha < 1
        else:
            alpha = 0

        if "nlevels" in kwargs:
            nlevels = kwargs["nlevels"]
            assert nlevels > 0
        else:
            # Single level is equivalent to Duffy transform
            nlevels = 1

        missing_measure = (alpha ** nlevels * source_box_extent) ** self.dim
        if abs(missing_measure) > 1e-6:
            from warnings import warn

            warn(
                "Droste probably has too few levels, missing measure = "
                + str(missing_measure)
            )

        if "result_array" in kwargs:
            result_array = kwargs["result_array"]

        else:
            result_array = self.make_result_array(**kwargs)

        # root brick
        root_brick = np.zeros((self.dim, 2))
        root_brick[:, 1] = source_box_extent

        # target points in 1D
        q_points = self.get_target_points(queue)
        assert len(q_points) == self.ntgt_points ** self.dim
        t = np.array([pt[-1] for pt in q_points[: self.ntgt_points]])

        # quad formula for each brick (normalized to [0,1])
        # sps.legendre blows up easily at high order
        import scipy.special as sps

        # legendre_nodes, _, legendre_weights = sps.legendre(
        #        nquad_points).weights.T
        legendre_nodes, legendre_weights = sps.p_roots(self.nquad_points)
        legendre_nodes = legendre_nodes * 0.5 + 0.5
        legendre_weights = legendre_weights * 0.5

        knl = self.get_cached_optimized_kernel()
        evt, res = knl(
            queue,
            alpha=alpha,
            result=result_array,
            root_brick=root_brick,
            target_nodes=t.astype(np.float64, copy=True),
            quadrature_nodes=legendre_nodes,
            quadrature_weights=legendre_weights,
            interaction_case_vecs=self.interaction_case_vecs.astype(
                np.float64, copy=True
            ),
            interaction_case_scls=self.interaction_case_scls.reshape(-1).astype(
                np.float64, copy=True
            ),
            n_cases=self.ncases,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )

        cheb_table = res["result"]

        return cheb_table

    def get_cache_key(self):
        return (
            type(self).__name__,
            str(self.dim) + "D",
            self.integral_knl.__str__(),
            "quad_order-" + str(self.ntgt_points),
            "brick_order-" + str(self.nquad_points),
            "brick_order-" + str(self.nquad_points),
        )

    def __call__(self, queue, **kwargs):
        """
        :arg source_box_extent
        :arg alpha
        :arg nlevels
        :arg extra_kernel_kwargs
        """
        if "cheb_coefs" in kwargs:
            assert len(kwargs["cheb_coefs"]) == self.n_q_points
            for ccoef in kwargs["cheb_coefs"]:
                assert len(ccoef) == self.nfunctions ** self.dim
            cheb_coefs = kwargs["cheb_coefs"]
            return self.postprocess_cheb_table(
                self.call_loopy_kernel(queue, **kwargs), cheb_coefs
            )
        else:
            logger.info("Returning cheb table directly")
            return self.call_loopy_kernel(queue, **kwargs)


# }}} End full Droste method

# {{{ reduced Droste method


class DrosteReduced(DrosteBase):
    """
    Reduce the workload by only building part of the table and infer the
    rest of the table by symmetry.
    """

    def __init__(
        self,
        integral_knl=None,
        quad_order=None,
        case_vecs=None,
        n_brick_quad_points=50,
        knl_symmetry_tags=None,
    ):
        super(DrosteReduced, self).__init__(
                integral_knl, quad_order, case_vecs, n_brick_quad_points)

        from volumential.list1_symmetry import CaseVecReduction

        # by default only self-interaction is counted
        if case_vecs is None:
            case_vecs = [np.array([0, 0, 0]), ]

        self.reduce_by_symmetry = CaseVecReduction(case_vecs, knl_symmetry_tags)
        logger.info(
            "Reduction ratio by symmetry = "
            + str(self.reduce_by_symmetry.get_full_reduction_ratio())
        )

        self.nbcases = len(self.reduce_by_symmetry.reduced_vecs)
        self.loop_domains = [None for bvec in self.reduce_by_symmetry.reduced_vecs]

        # state of the object that determines behavior of functions like
        # get_kernel()
        # get_kernel() returns the kernel for actual quadrature when
        #   get_kernel_id = 0,
        # while it returns the kernel for expansion within the same case when = 1.
        self.current_base_case = 0
        self.get_kernel_id = 0

    def make_loop_domain(self, base_case_id):
        # The icase is just a dummy variable that allows resusing some of the
        # code (e.g. kernel evals) from the base class
        assert base_case_id >= 0
        assert base_case_id < self.nbcases

        tgt_vars = self.make_tgt_vars()  # noqa: F841
        quad_vars = self.make_quad_vars()
        basis_vars = self.make_basis_vars()
        basis_eval_vars = self.make_basis_eval_vars()
        pwaffs = self.make_pwaffs()
        loop_domain_common_parts = (
            self.make_brick_domain(quad_vars, self.nquad_points)
            & self.make_brick_domain(basis_vars, self.nfunctions)
            & self.make_brick_domain(basis_eval_vars, self.nfunctions)
            & self.make_brick_domain(["iside"], 2)
            & self.make_brick_domain(["iaxis"], self.dim)
            & self.make_brick_domain(["ibrick_side"], 2)
            & self.make_brick_domain(["ibrick_axis"], self.dim)
            & self.make_brick_domain(["ilevel"], "nlevels")
            & self.make_brick_domain(["icase"], 1)
        )

        # target brick domain depends on symmetry group of the case vec
        # upper bounds account for flippings, and lower bounds account for swappings
        flippable, swappable_groups = self.reduce_by_symmetry.parse_symmetry_tags(
            self.reduce_by_symmetry.reduced_invariant_groups[base_case_id]
        )
        prev_swappable = -np.ones(self.dim, dtype=np.int32)
        for iaxis in range(self.dim):
            for group in swappable_groups:
                if iaxis in group:
                    axid = list(group).index(iaxis)
                    if axid == 0:
                        pass
                    else:
                        prev_swappable[iaxis] = list(group)[axid - 1]

        from functools import reduce
        import operator

        tgt_domain_ubounds = reduce(
            operator.and_,
            [
                pwaffs["t" + str(iaxis)].lt_set(
                    pwaffs[0]
                    + (
                        (self.ntgt_points + 1) // 2
                        if flippable[iaxis]
                        else self.ntgt_points
                    )
                )
                for iaxis in range(self.dim)
            ],
        )
        tgt_domain_lbounds = reduce(
            operator.and_,
            [
                (
                    pwaffs[0]
                    if prev_swappable[iaxis] == -1
                    else pwaffs["t" + str(prev_swappable[iaxis])]
                ).le_set(pwaffs["t" + str(iaxis)])
                for iaxis in range(self.dim)
            ],
        )
        tgt_domain = tgt_domain_ubounds & tgt_domain_lbounds

        self.loop_domains[base_case_id] = loop_domain_common_parts & tgt_domain
        return self.loop_domains[base_case_id]

    def make_result_array(self, **kwargs):
        """Allocate memory space for results.
        """
        # by default uses double type returns
        if "result_dtype" in kwargs:
            result_dtype = kwargs["result_dtype"]
        else:
            result_dtype = np.float64

        # allocate return arrays
        if self.dim == 1:
            result_array = (
                    np.zeros(
                        (self.nfunctions, self.ntgt_points, 1),
                        result_dtype
                        )
                    + np.nan)
        elif self.dim == 2:
            result_array = (
                    np.zeros(
                        (
                            self.nfunctions,
                            self.nfunctions,
                            self.ntgt_points,
                            self.ntgt_points,
                            1,
                            ),
                        result_dtype
                        )
                    + np.nan
                    )
        elif self.dim == 3:
            result_array = (
                    np.zeros(
                        (
                            self.nfunctions,
                            self.nfunctions,
                            self.nfunctions,
                            self.ntgt_points,
                            self.ntgt_points,
                            self.ntgt_points,
                            1,
                            ),
                        result_dtype
                        )
                    + np.nan
                    )
        else:
            raise NotImplementedError
        return result_array

    def get_kernel_expansion_by_symmetry_code(self):
        """
        Extra assignments that performs expansion by symmetry within the
        current case.
        """
        # case_id = self.reduce_by_symmetry.reduced_vec_ids[self.current_base_case]
        # case_vec = self.reduce_by_symmetry.reduced_vecs[self.current_base_case]
        invariant_group = self.reduce_by_symmetry.reduced_invariant_groups[
            self.current_base_case
        ]

        flippable, swappable_groups = self.reduce_by_symmetry.parse_symmetry_tags(
            invariant_group
        )
        nflippables = int(sum(flippable))
        assert len(flippable) == self.dim
        flippable_ids = [i for i in range(self.dim) if flippable[i]]
        base_tgt_ordering = self.make_dim_independent("TGT_VARS").split(",")
        base_fun_ordering = self.make_dim_independent("BASIS_VARS").split(",")
        assert len(base_tgt_ordering) == self.dim

        def flip_tgt(tgt_var):
            assert isinstance(tgt_var, str)
            return str(self.ntgt_points) + " - " + tgt_var + " - 1"

        def conjunct(var):
            # conjunct(v_i) = v_{dim-1-i}
            assert isinstance(var, str)
            vid = int(var[1:])
            assert vid >= 0 and vid < self.dim
            return var[0] + str(self.dim - 1 - vid)

        from itertools import product, permutations

        ext_ids = [
            ext
            for ext in product(
                [fid for fid in product([0, 1], repeat=nflippables)],
                *[
                    [sid for sid in permutations(sgroup)]
                    for sgroup in swappable_groups
                    ]
            )
        ]

        # print(ext_ids)

        # The idea is that, any member of the hyperoctahedral group
        # has the decomposition m = f * s, where f is a flip and s
        # is a swap.
        # Conversely, by iterating all combinations of flips and swaps,
        # the whole group is iterated over.

        expansion_code = []
        for ext_index, ext_id in zip(range(len(ext_ids)), ext_ids):
            ext_tgt_ordering = list(base_tgt_ordering)
            ext_fun_ordering = list(base_fun_ordering)
            # apply s
            for swap in ext_id[1:]:
                ns = len(swap)
                for i in range(ns):
                    original_part = sorted(swap)
                    ext_tgt_ordering[original_part[i]] = base_tgt_ordering[swap[i]]
                    ext_fun_ordering[original_part[i]] = base_fun_ordering[swap[i]]
            # apply f, also figure out the rule for sign changes
            ext_sign = "1"
            for fid in range(nflippables):
                if ext_id[0][fid]:
                    iaxis = flippable_ids[fid]
                    itgt = ext_tgt_ordering.index("t" + str(iaxis))
                    ext_tgt_ordering[itgt] = flip_tgt(ext_tgt_ordering[itgt])
                    ext_sign = (
                        ext_sign
                        + " * if("
                        + self.basis_vars[iaxis]
                        + " % 2 == 0, 1, -1)"
                    )

            # print(', '.join(ext_tgt_ordering))
            ext_tgt_ordering.reverse()
            ext_fun_ordering.reverse()

            ext_instruction_ids = ':'.join([
                'result_ext_' + str(eid)
                for eid in range(len(ext_ids))
                if eid != ext_index
                ])

            expansion_code += [
                self.make_dim_independent(
                    """
                for BASIS_VARS, TGT_VARS, icase
                    result[EXT_BASIS_VARS, EXT_TGT_VARS, icase] = (
                        EXT_SIGN *
                        result[REV_BASIS_VARS, REV_TGT_VARS, icase]
                        ) {id=result_ext_EXT_ID,nosync=EXT_INSN_IDS}
                end
                """.replace(
                        "EXT_TGT_VARS", ",".join(ext_tgt_ordering)
                    )
                    .replace("EXT_BASIS_VARS", ",".join(ext_fun_ordering))
                    .replace("EXT_SIGN", ext_sign)
                    .replace("EXT_ID", str(ext_index))
                    .replace("EXT_INSN_IDS", ext_instruction_ids)
                )
            ]

        return expansion_code

    def get_kernel(self, **kwargs):
        """Reduced Droste is a 2-staged algorithm. The first stage uses the
        kernel from DrosteBase to build part of the table. In the second
        stage, an expansion kernel is called to fill the empty entries.
        """

        domain = self.make_loop_domain(base_case_id=self.current_base_case)

        extra_kernel_kwarg_types = ()
        if "extra_kernel_kwarg_types" in kwargs:
            extra_kernel_kwarg_types = kwargs["extra_kernel_kwarg_types"]

        if self.get_kernel_id == 0:
            loopy_knl = lp.make_kernel(  # NOQA
                [domain],
                self.get_kernel_code()
                # FIXME: cannot have expansion in the same kernel, since it
                # will require a global barrier
                # + self.get_kernel_expansion_by_symmetry_code()
                + self.get_sumpy_kernel_eval_insns(),
                [
                    lp.ValueArg("alpha", np.float64),
                    lp.ValueArg("n_cases, nfunctions, quad_order, dim", np.int32),
                    lp.GlobalArg("interaction_case_vecs",
                        np.float64, "dim, n_cases"),
                    lp.GlobalArg("interaction_case_scls", np.float64, "n_cases"),
                    lp.GlobalArg("target_nodes", np.float64, "quad_order"),
                    lp.GlobalArg(
                        "result", None,
                        ", ".join(
                            ["nfunctions" for d in range(self.dim)]
                            + ["quad_order" for d in range(self.dim)]
                        )
                        + ", n_cases",
                    ), ] + list(extra_kernel_kwarg_types)
                    + ["...", ],
                name="brick_map",
                lang_version=(2018, 2),
            )

        elif self.get_kernel_id == 1:
            loopy_knl = lp.make_kernel(  # NOQA
                [domain],
                self.get_kernel_expansion_by_symmetry_code(),
                [
                    lp.ValueArg("n_cases, nfunctions, quad_order, dim", np.int32),
                    lp.GlobalArg(
                        "result", None,
                        ", ".join(
                            ["nfunctions" for d in range(self.dim)]
                            + ["quad_order" for d in range(self.dim)]
                        )
                        + ", n_cases",
                    ), ] + list(extra_kernel_kwarg_types)
                    + ["...", ],
                name="brick_map_expansion",
                lang_version=(2018, 2),
            )

        else:
            raise NotImplementedError

        loopy_knl = lp.fix_parameters(loopy_knl, dim=self.dim)
        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        try:
            loopy_knl = self.integral_knl.prepare_loopy_kernel(loopy_knl)
        except Exception:
            pass

        return loopy_knl

    def get_optimized_kernel(self, ncpus=None, **kwargs):
        # The returned kernel depends on the state variable
        # self.current_base_case, self.get_kernel_id
        if ncpus is None:
            import multiprocessing
            # NOTE: this detects the number of logical cores, which
            # may result in suboptimal performance.
            ncpus = multiprocessing.cpu_count()

        knl = self.get_kernel(**kwargs)
        knl = lp.join_inames(knl, inames=self.basis_vars, new_iname="func")
        knl = lp.split_iname(knl, "func", ncpus, inner_tag="g.0")
        return knl

    def call_loopy_kernel_case(self, queue, base_case_id, **kwargs):
        """
        Call the table builder on one base case, as given in :self.current_base_case:
        :arg source_box_extent
        :arg alpha
        :arg nlevels
        :arg extra_kernel_kwargs
        """

        if base_case_id != self.current_base_case:
            self.current_base_case = base_case_id

        if "source_box_extent" in kwargs:
            assert kwargs["source_box_extent"] > 0
            source_box_extent = kwargs["source_box_extent"]
        else:
            source_box_extent = 1

        extra_kernel_kwargs = {}
        if "extra_kernel_kwargs" in kwargs:
            extra_kernel_kwargs = kwargs["extra_kernel_kwargs"]

        if "alpha" in kwargs:
            alpha = kwargs["alpha"]
            assert alpha >= 0 and alpha < 1
        else:
            alpha = 0

        if "nlevels" in kwargs:
            nlevels = kwargs["nlevels"]
            assert nlevels > 0
        else:
            # Single level is equivalent to Duffy transform
            nlevels = 1

        missing_measure = (alpha ** nlevels * source_box_extent) ** self.dim
        if abs(missing_measure) > 1e-6:

            logger.warn(
                "Droste probably has too few levels, missing measure = %e"
                % missing_measure
            )

        if "result_array" in kwargs:
            result_array = kwargs["result_array"]
        else:
            result_array = self.make_result_array(**kwargs)

        # root brick
        root_brick = np.zeros((self.dim, 2))
        root_brick[:, 1] = source_box_extent

        # target points in 1D
        q_points = self.get_target_points(queue)
        assert len(q_points) == self.ntgt_points ** self.dim
        t = np.array([pt[-1] for pt in q_points[: self.ntgt_points]])

        base_case_vec = np.array(
            [
                [self.reduce_by_symmetry.reduced_vecs[self.current_base_case][d]]
                for d in range(self.dim)
            ]
        )
        base_case_scl = np.array(
            [
                self.interaction_case_scls[
                    self.reduce_by_symmetry.reduced_vec_ids[self.current_base_case]
                ]
            ]
        )

        # quad formula for each brick (normalized to [0,1])
        # sps.legendre blows up easily at high order
        import scipy.special as sps

        # legendre_nodes, _, legendre_weights = sps.legendre(
        #        nquad_points).weights.T
        legendre_nodes, legendre_weights = sps.p_roots(self.nquad_points)
        legendre_nodes = legendre_nodes * 0.5 + 0.5
        legendre_weights = legendre_weights * 0.5

        self.get_kernel_id = 0
        try:
            delattr(self, "_memoize_dic_get_cached_optimized_kernel")
        except Exception:
            pass
        knl = self.get_cached_optimized_kernel()
        evt, res = knl(
            queue,
            alpha=alpha,
            result=result_array,
            root_brick=root_brick,
            target_nodes=t.astype(np.float64, copy=True),
            quadrature_nodes=legendre_nodes,
            quadrature_weights=legendre_weights,
            interaction_case_vecs=base_case_vec.astype(np.float64, copy=True),
            interaction_case_scls=base_case_scl.astype(np.float64, copy=True),
            n_cases=1,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )
        raw_cheb_table_case = res["result"]

        self.get_kernel_id = 1
        try:
            delattr(self, "_memoize_dic_get_cached_optimized_kernel")
        except Exception:
            pass
        knl2 = self.get_cached_optimized_kernel()
        evt, res2 = knl2(
            queue,
            result=raw_cheb_table_case,
            n_cases=1,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )
        cheb_table_case = res2["result"]

        return cheb_table_case

    def build_cheb_table(self, queue, **kwargs):
        # Build the table using Cheb modes as sources
        if self.dim == 1:
            cheb_table = (
                np.zeros((self.nfunctions, self.ntgt_points, self.ncases)) + np.nan
            )
        elif self.dim == 2:
            cheb_table = (
                np.zeros(
                    (
                        self.nfunctions,
                        self.nfunctions,
                        self.ntgt_points,
                        self.ntgt_points,
                        self.ncases,
                    )
                )
                + np.nan
            )
        elif self.dim == 3:
            cheb_table = (
                np.zeros(
                    (
                        self.nfunctions,
                        self.nfunctions,
                        self.nfunctions,
                        self.ntgt_points,
                        self.ntgt_points,
                        self.ntgt_points,
                        self.ncases,
                    )
                )
                + np.nan
            )
        else:
            raise NotImplementedError

        for base_case_id in range(self.nbcases):
            # print(base_case_id)
            self.current_base_case = base_case_id
            case_id = self.reduce_by_symmetry.reduced_vec_ids[base_case_id]
            cheb_table[..., case_id] = self.call_loopy_kernel_case(
                queue, base_case_id, **kwargs
            )[..., 0]
            # print("Loopy kernel finished")

            # {{{ expansion by symmetry
            flippable, swappable_groups = \
                    self.reduce_by_symmetry.parse_symmetry_tags(
                            self.reduce_by_symmetry.symmetry_tags
                            )
            nflippables = int(sum(flippable))
            assert len(flippable) == self.dim
            flippable_ids = [i for i in range(self.dim) if flippable[i]]

            from itertools import product, permutations

            ext_ids = [
                ext
                for ext in product(
                    [fid for fid in product([0, 1], repeat=nflippables)],
                    *[
                        [sid for sid in permutations(sgroup)]
                        for sgroup in swappable_groups
                    ]
                )
            ]

            base_case_vec = self.reduce_by_symmetry.full_vecs[case_id]
            assert (base_case_vec
                    == self.reduce_by_symmetry.reduced_vecs[base_case_id])
            for ext_index, ext_id in zip(range(len(ext_ids)), ext_ids):
                # start from the base vec
                ext_vec = list(base_case_vec)
                swapped_axes = list(range(self.dim))

                # apply s
                # Note that fi and ti are in reverse order
                for swap in ext_id[1:]:
                    ns = len(swap)
                    for i in range(ns):
                        original_part = sorted(swap)
                        ext_vec[original_part[i]] = base_case_vec[swap[i]]
                        swapped_axes[self.dim - 1 - original_part[i]] = (
                            self.dim - 1 - swap[i]
                        )
                swapped_basis_axes = swapped_axes
                swapped_tgt_axes = [ax + self.dim for ax in swapped_basis_axes]
                tmp_table_part = (
                    cheb_table[..., case_id]
                    .transpose(swapped_basis_axes + swapped_tgt_axes)
                    .copy()
                )

                # if case_id == 69:
                #    print("after swapping:",
                #           np.linalg.norm(tmp_table_part - cheb_table[...,case_id]))

                # apply f, also figure out the rule for sign changes
                for fid in range(nflippables):
                    if ext_id[0][fid]:
                        iaxis = flippable_ids[fid]

                        if 0:  # ext_vec[swapped_iaxis] == 0:
                            continue
                        else:
                            ext_vec[iaxis] = -ext_vec[iaxis]
                            # flips.append(
                            #   lambda x:
                            #     np.flip(x, (self.dim - 1 - iaxis)
                            #     + self.dim))
                            full_slice = [
                                slice(None, None) for axid in range(2 * self.dim)
                            ]
                            for basis_id in range(self.nfunctions):
                                if basis_id % 2 == 0:
                                    continue
                                sign_slice = full_slice
                                sign_slice[self.dim - 1 - iaxis] = basis_id
                                np.multiply.at(tmp_table_part, tuple(sign_slice), -1)

                        tmp_table_part = np.flip(
                            tmp_table_part, (self.dim - 1 - iaxis) + self.dim
                        )

                # if case_id == 69:
                #    print("after flipping:",
                #       np.linalg.norm(tmp_table_part - cheb_table[...,case_id]))
                #    print("Filling by symmetry",
                #       base_case_vec, "-->", ext_vec, swapped_axes, ext_id)

                assert tuple(ext_vec) in self.reduce_by_symmetry.full_vecs
                ext_case_id = self.reduce_by_symmetry.full_vecs.index(tuple(ext_vec))
                cheb_table[..., ext_case_id] = tmp_table_part

                # }}} End expansion by symmetry

        return cheb_table

    def get_cache_key(self):
        if self.reduce_by_symmetry.symmetry_tags is None:
            # Bn: the full n-dimensional hyperoctahedral group
            symmetry_info = 'B%d' % self.dim
        else:
            symmetry_info = "Span{%s}" % ",".join([
                repr(tag) for tag in
                sorted(self.reduce_by_symmetry.symmetry_tags)
                ])

        return (
            type(self).__name__,
            str(self.dim) + "D",
            self.integral_knl.__str__(),
            "quad_order-" + str(self.ntgt_points),
            "brick_order-" + str(self.nquad_points),
            "case-" + str(self.current_base_case),
            "kernel_id-" + str(self.get_kernel_id),
            "symmetry-" + symmetry_info,
        )

    def __call__(self, queue, **kwargs):
        """
        :arg source_box_extent
        :arg alpha
        :arg nlevels
        :arg extra_kernel_kwargs
        """
        if "cheb_coefs" in kwargs:
            assert len(kwargs["cheb_coefs"]) == self.n_q_points
            for ccoef in kwargs["cheb_coefs"]:
                assert len(ccoef) == self.nfunctions ** self.dim
            cheb_coefs = kwargs["cheb_coefs"]
            return self.postprocess_cheb_table(
                self.build_cheb_table(queue, **kwargs), cheb_coefs
            )
        else:
            logger.info("Returning cheb table directly")
            return self.build_cheb_table(queue, **kwargs)


# }}} End reduced Droste method

# {{{ inverse Droste method


class InverseDrosteReduced(DrosteReduced):
    r"""
    A variant of the Droste method.
    Instead of computing a volume potential, it computes the "inverse" to a
    Riesz potential, aka a "fractional Laplacian".

    Specifically, given an integral kernel G(r), this class supports the
    precomputation for integrals of the form

    .. math::

       \int_B G(r) (u(x) - u(y)) dy

    For k-dimensional fractional Laplacian, :math:`G(r) = \frac{1}{r^{k+2s}}`.

    The core part of concern (that is to be modified based on DrosteReduces):

    .. code-block::

        ...

        for BASIS_VARS, TGT_VARS, icase
            for ilevel
                for ibrick_axis, ibrick_side, QUAD_VARS

                    PREPARE_BASIS_VALS

                    <> density_val = DENSITY_VAL_ASSIGNMENT \
                            {id=density,dep=basis_evals}

                end
            end
        end

        ...

    """

    def __init__(
            self,
            integral_knl,
            quad_order,
            case_vecs,
            n_brick_quad_points=50,
            knl_symmetry_tags=None,
            auto_windowing=True):
        """
        :param auto_windowing: auto-detect window radius.
        """
        super(InverseDrosteReduced, self).__init__(
            integral_knl, quad_order, case_vecs,
            n_brick_quad_points, knl_symmetry_tags)
        self.auto_windowing = auto_windowing

    def get_cache_key(self):
        return (
            type(self).__name__,
            str(self.dim) + "D",
            self.integral_knl.__str__(),
            "quad_order-" + str(self.ntgt_points),
            "brick_order-" + str(self.nquad_points),
            "case-" + str(self.current_base_case),
            "kernel_id-" + str(self.get_kernel_id),
        )

    def codegen_basis_tgt_eval(self, iaxis):
        """Generate instructions to evaluate Chebyshev polynomial basis
        at the target point, given that the target point lies in the
        source box. (Chebyshev polynomials of the first kind T_n).

        If the target point is not in the source box, the concerned instructions
        will return 0.
        """

        # only valid for self-interactions
        assert all([self.reduce_by_symmetry.reduced_vecs[
            self.current_base_case][d] == 0 for d in range(self.dim)])

        code = """  # noqa
            <> T0_tgt_IAXIS = 1
            <> T1_tgt_IAXIS = template_true_target[IAXIS] {dep=template_true_targets}
            <> Tprev_tgt_IAXIS = T0_tgt_IAXIS {id=t0_tgt_IAXIS}
            <> Tcur_tgt_IAXIS = T1_tgt_IAXIS {id=t1_tgt_IAXIS,dep=t0_tgt_IAXIS}

            for pIAXIS
                <> Tnext_tgt_IAXIS = (2 * template_true_target[IAXIS] * Tcur_tgt_IAXIS
                                - Tprev_tgt_IAXIS) {id=tnext_tgt_IAXIS,dep=t1_tgt_IAXIS}
                Tprev_tgt_IAXIS = Tcur_tgt_IAXIS {id=tprev_tgt_updateIAXIS,dep=tnext_tgt_IAXIS}
                Tcur_tgt_IAXIS = Tnext_tgt_IAXIS {id=tcur_tgt_updateIAXIS,dep=tprev_tgt_updateIAXIS}
            end

            <> basis_tgt_evalIAXIS = (
                T0_tgt_IAXIS * if(fIAXIS == 0, 1, 0)
                + T1_tgt_IAXIS * if(fIAXIS == 1, 1, 0)
                + simul_reduce(sum, pIAXIS,
                    if(fIAXIS >= 2 and fIAXIS == pIAXIS + 2, Tcur_tgt_IAXIS, 0))

                ) {id=tgtbasisIAXIS,dep=tcur_tgt_updateIAXIS}
            """.replace(
            "IAXIS", str(iaxis)
        )
        return code

    def codegen_der2_basis_tgt_eval(self, iaxis):
        r"""Generate instructions to evaluate the second order derivatives
        of Chebyshev polynomial basis at the target point, given that the target
        lies in the source box. (Chebyshev polynomials of the first kind T_n).

        If the target point is not in the source box, the concerned instructions
        will return 0.

        The evaluation is based on Chebyshev polynomials of the second kind
        :math:`U_n`.

        .. math::

           \frac{d^2 T_n}{dx^2} = n \frac{(n+1)T_n - U_n}{x^2 - 1}
        """

        # only valid for self-interactions
        assert all([self.reduce_by_symmetry.reduced_vecs[
            self.current_base_case][d] == 0 for d in range(self.dim)])

        code = """  # noqa
            <> U0_tgt_IAXIS = 1
            <> U1_tgt_IAXIS = 2 * template_true_target[IAXIS] {dep=template_true_targets}
            <> Uprev_tgt_IAXIS = U0_tgt_IAXIS {id=u0_tgt_IAXIS}
            <> Ucur_tgt_IAXIS = U1_tgt_IAXIS {id=u1_tgt_IAXIS,dep=u0_tgt_IAXIS}

            for pIAXIS
                <> Unext_tgt_IAXIS = (2 * template_true_target[IAXIS] * Ucur_tgt_IAXIS
                                - Uprev_tgt_IAXIS) {id=unext_tgt_IAXIS,dep=u1_tgt_IAXIS}
                Uprev_tgt_IAXIS = Ucur_tgt_IAXIS {id=uprev_tgt_updateIAXIS,dep=unext_tgt_IAXIS}
                Ucur_tgt_IAXIS = Unext_tgt_IAXIS {id=ucur_tgt_updateIAXIS,dep=uprev_tgt_updateIAXIS}
            end

            # U_n(target)
            <> basis2_tgt_evalIAXIS = (
                U0_tgt_IAXIS * if(fIAXIS == 0, 1, 0)
                + U1_tgt_IAXIS * if(fIAXIS == 1, 1, 0)
                + simul_reduce(sum, pIAXIS,
                    if(fIAXIS >= 2 and fIAXIS == pIAXIS + 2, Ucur_tgt_IAXIS, 0))
                ) {id=tgtbasis2IAXIS,dep=ucur_tgt_updateIAXIS}

            # this temp var helps with type deduction
            <> f_order_IAXIS = fIAXIS
            <> der2_basis_tgt_evalIAXIS = f_order_IAXIS * (
                    ((f_order_IAXIS + 1) * basis_tgt_evalIAXIS - basis2_tgt_evalIAXIS)
                    / (template_true_target[IAXIS]**2 - 1)
                ) * (2**2) / (root_extent[IAXIS]**2) {id=tgtd2basisIAXIS,dep=tgtbasisIAXIS:tgtbasis2IAXIS}
            """.replace(
                    "IAXIS", str(iaxis)
                    )
        return code

    def codegen_windowing_function(self):
        r"""Given :math:`dist = x - y`, compute the windowing function.
        """

        code = []
        if self.dim == 1:
            code.append("<> distsq = dist[0]*dist[0]")
        elif self.dim == 2:
            code.append("<> distsq = dist[0]*dist[0] + dist[1]*dist[1]")
        elif self.dim == 3:
            code.append("<> distsq = dist[0]*dist[0] + \
                                     dist[1]*dist[1] + dist[2]*dist[2]")
        # renormalized distance
        code.append("<> rndist = sqrt(distsq) / delta")

        if True:
            # polynomial windowing
            logger.info("Using polynomial windowing function.")
            code.append(r"""
                <> windowing = if(rndist >= 1,
                                  0,
                                  (1 - 35 * (rndist**4)
                                  + 84 * (rndist**5)
                                  - 70 * (rndist**6)
                                  + 20 * (rndist**7)
                                  )
                                 )
                """)
        elif False:
            # classical bump function
            logger.info("Using bump windowing function.")
            code.append(
                "<> windowing = exp(1) * exp(-(1/(1 - (distsq / (delta*delta)))))")
        else:
            # smooth transitions of 0-->1-->0
            logger.info("Using smooth transition windowing function.")
            code.append("<> fv = if(rndist > 0, exp(-1 / rndist), 0)")
            code.append("<> fc = if(1 - rndist > 0, exp(-1 / (1 - rndist)), 0)")
            code.append("<> windowing = 1 - fv / (fv + fc)")

        return '\n'.join(code)

    def make_dim_independent(self, knlstring):
        r"""Produce the correct
        :math:`\text{DENSITY_VAL_ASSIGNMENT} = u(x) - u(y)`
        for self-interactions.
        """

        # detect for self-interactions
        if all([self.reduce_by_symmetry.reduced_vecs[
                self.current_base_case][d] == 0 for d in range(self.dim)]):
            target_box_is_source = True
        else:
            target_box_is_source = False

        # replace REV_* first
        resknl = knlstring.replace("REV_BASIS_VARS", ",".join(self.basis_vars[::-1]))
        resknl = resknl.replace("REV_TGT_VARS", ",".join(self.tgt_vars[::-1]))

        resknl = resknl.replace("BASIS_VARS", ",".join(self.basis_vars))
        resknl = resknl.replace("TGT_VARS", ",".join(self.tgt_vars))
        resknl = resknl.replace("QUAD_VARS", ",".join(self.quad_vars))

        if self.get_kernel_id == 0:
            resknl = resknl.replace(
                    "POSTPROCESS_KNL_VAL",
                    '\n'.join([
                        self.codegen_windowing_function(),
                        "<> knl_val_post = windowing * knl_val {id=pp_kval}"
                        ])
                    )
        elif self.get_kernel_id == 1:
            resknl = resknl.replace(
                    "POSTPROCESS_KNL_VAL",
                    '\n'.join([
                        self.codegen_windowing_function(),
                        "<> knl_val_post = (1 - windowing) * knl_val {id=pp_kval}"
                        ])
                    )
        else:
            pass

        # {{{ density evals

        basis_eval_insns = [self.codegen_basis_eval(i) for i in range(self.dim)]

        if target_box_is_source:
            basis_eval_insns += [
                    self.codegen_basis_tgt_eval(i) for i in range(self.dim)]

        if self.get_kernel_id == 0:
            # Given target x,
            # u(x) - u(y) p.v. integrated around a small region symmetric to x,
            # truncated to second order 0.5 * [(x - y)' * diag(Hess(u)(x)) * (x - y)]
            if target_box_is_source:
                basis_eval_insns += [
                        self.codegen_der2_basis_tgt_eval(i) for i in range(self.dim)]

                resknl = resknl.replace(
                        "PREPARE_BASIS_VALS",
                        "\n".join(basis_eval_insns + [
                            "... nop {id=basis_evals,dep=%s}"
                            % ':'.join(
                                ['basis%d' % i for i in range(self.dim)]
                                + ['tgtbasis%d' % i for i in range(self.dim)]
                                + ['tgtd2basis%d' % i for i in range(self.dim)]
                                ),
                            ])
                        )
            else:
                resknl = resknl.replace(
                        "PREPARE_BASIS_VALS",
                        "\n".join(basis_eval_insns + [
                            "... nop {id=basis_evals,dep=%s}"
                            % ':'.join(
                                ['basis%d' % i for i in range(self.dim)]
                                ),
                            ])
                        )

            if self.dim == 1:
                if target_box_is_source:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            ' '.join([
                                "0.5 * der2_basis_tgt_eval0 * (dist[0]**2)",
                                ])
                            )
                else:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            "- basis_eval0"
                            )

            elif self.dim == 2:
                if target_box_is_source:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            ' '.join([
                                "  0.5 * der2_basis_tgt_eval0 * basis_tgt_eval1 * (dist[0]**2)",  # noqa: E501
                                "+ 0.5 * basis_tgt_eval0 * der2_basis_tgt_eval1 * (dist[1]**2)",  # noqa: E501
                                ])
                            )
                else:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            "- basis_eval0 * basis_eval1"
                            )

            elif self.dim == 3:
                if target_box_is_source:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            ' '.join([
                                "  0.5 * der2_basis_tgt_eval0 * basis_tgt_eval1 * basis_tgt_eval2 * (dist[0]**2)",  # noqa: E501
                                "+ 0.5 * basis_tgt_eval0 * der2_basis_tgt_eval1 * basis_tgt_eval2 * (dist[1]**2)",  # noqa: E501
                                "+ 0.5 * basis_tgt_eval0 * basis_tgt_eval1 * der2_basis_tgt_eval2 * (dist[2]**2)",  # noqa: E501
                                ])
                            )
                else:
                    resknl = resknl.replace(
                            "DENSITY_VAL_ASSIGNMENT",
                            "- basis_eval0 * basis_eval1 * basis_eval2"
                            )

            else:  # self.dim not in [1, 2, 3]
                raise NotImplementedError("No support for dimension %d" % self.dim)

        elif self.get_kernel_id == 1:

            if target_box_is_source:
                # u(x) - u(y)
                resknl = resknl.replace(
                        "PREPARE_BASIS_VALS",
                        "\n".join(basis_eval_insns + [
                            "... nop {id=basis_evals,dep=%s}"
                            % ':'.join(
                                ['basis%d' % i for i in range(self.dim)]
                                + ['tgtbasis%d' % i for i in range(self.dim)]
                                ),
                            ])
                        )
                resknl = resknl.replace(
                        "DENSITY_VAL_ASSIGNMENT",
                        ' - '.join([
                            ' * '.join(
                                ["basis_tgt_eval%d" % i for i in range(self.dim)]),
                            ' * '.join(
                                ["basis_eval%d" % i for i in range(self.dim)]),
                            ])
                        )
            else:
                # - u(y)
                resknl = resknl.replace(
                        "PREPARE_BASIS_VALS",
                        "\n".join(basis_eval_insns + [
                            "... nop {id=basis_evals,dep=%s}"
                            % ':'.join(
                                ['basis%d' % i for i in range(self.dim)]
                                ),
                            ])
                        )
                resknl = resknl.replace(
                        "DENSITY_VAL_ASSIGNMENT",
                        ' - ' + ' * '.join(
                            ["basis_eval%d" % i for i in range(self.dim)]
                            )
                        )

        # }}} End density evals

        resknl = resknl.replace(
            "PROD_QUAD_WEIGHT",
            " * ".join(
                [
                    "quadrature_weights[QID]".replace("QID", qvar)
                    for qvar in self.quad_vars
                ]
            ),
        )

        if self.dim == 1:
            resknl = resknl.replace("TPLTGT_ASSIGNMENT", """target_nodes[t0]""")
            resknl = resknl.replace("QUAD_PT_ASSIGNMENT", """quadrature_nodes[q0]""")

        elif self.dim == 2:
            resknl = resknl.replace(
                "TPLTGT_ASSIGNMENT",
                """if(iaxis == 0, target_nodes[t0], target_nodes[t1])""",
            )
            resknl = resknl.replace(
                "QUAD_PT_ASSIGNMENT",
                """if(iaxis == 0, quadrature_nodes[q0], quadrature_nodes[q1])""",
            )

        elif self.dim == 3:
            resknl = resknl.replace(
                "TPLTGT_ASSIGNMENT",
                """if(iaxis == 0, target_nodes[t0], if(
                          iaxis == 1, target_nodes[t1], target_nodes[t2]))""",
            )
            resknl = resknl.replace(
                "QUAD_PT_ASSIGNMENT",
                """if(iaxis == 0, quadrature_nodes[q0], if(
                  iaxis == 1, quadrature_nodes[q1], quadrature_nodes[q2]))""",
            )

        else:
            raise NotImplementedError

        return resknl

    def get_kernel(self, **kwargs):
        """Get loopy kernel for computation, the get_kernel_id determines
        what task to perform.

        - 0:    Integrate :math:`W(r) G(r) [u(x) - u(y) + grad(u)(y - x)]`
        - 1:    Add the integral of :math:`[1 - W(r)] G(r) [u(x) - u(y)]`
        - 2:    Expansion by symmetry.
        """
        domain = self.make_loop_domain(base_case_id=self.current_base_case)

        extra_kernel_kwarg_types = ()
        if "extra_kernel_kwarg_types" in kwargs:
            extra_kernel_kwarg_types = kwargs["extra_kernel_kwarg_types"]

        extra_loopy_kernel_kwargs = {}
        if "extra_loopy_kernel_kwargs" in kwargs:
            extra_loopy_kernel_kwargs = kwargs["extra_loopy_kernel_kwargs"]

        if self.get_kernel_id == 0 or self.get_kernel_id == 1:
            loopy_knl = lp.make_kernel(  # NOQA
                [domain],
                self.get_kernel_code()
                + self.get_sumpy_kernel_eval_insns(),
                [
                    lp.ValueArg("alpha", np.float64),
                    lp.ValueArg("delta", np.float64),
                    lp.ValueArg("n_cases, nfunctions, quad_order, dim", np.int32),
                    lp.GlobalArg("interaction_case_vecs",
                        np.float64, "dim, n_cases"),
                    lp.GlobalArg("interaction_case_scls", np.float64, "n_cases"),
                    lp.GlobalArg("target_nodes", np.float64, "quad_order"),
                    lp.GlobalArg(
                        "result", None,
                        ", ".join(
                            ["nfunctions" for d in range(self.dim)]
                            + ["quad_order" for d in range(self.dim)]
                        )
                        + ", n_cases",
                    ), ] + list(extra_kernel_kwarg_types)
                    + ["...", ],
                name="brick_map_%d" % self.get_kernel_id,
                lang_version=(2018, 2),
                **extra_loopy_kernel_kwargs
            )

        elif self.get_kernel_id == 2:
            loopy_knl = lp.make_kernel(  # NOQA
                [domain],
                self.get_kernel_expansion_by_symmetry_code(),
                [
                    lp.ValueArg("n_cases, nfunctions, quad_order, dim", np.int32),
                    lp.GlobalArg(
                        "result", None,
                        ", ".join(
                            ["nfunctions" for d in range(self.dim)]
                            + ["quad_order" for d in range(self.dim)]
                        )
                        + ", n_cases",
                    ), ] + list(extra_kernel_kwarg_types)
                    + ["...", ],
                name="brick_map_expansion",
                lang_version=(2018, 2),
                **extra_loopy_kernel_kwargs
            )

        else:
            raise NotImplementedError

        loopy_knl = lp.fix_parameters(loopy_knl, dim=self.dim)
        loopy_knl = lp.set_options(loopy_knl, write_cl=False)
        loopy_knl = lp.set_options(loopy_knl, return_dict=True)

        # loopy_knl = lp.make_reduction_inames_unique(loopy_knl)

        try:
            loopy_knl = self.integral_knl.prepare_loopy_kernel(loopy_knl)
        except Exception:
            pass

        return loopy_knl

    def call_loopy_kernel_case(self, queue, base_case_id, **kwargs):
        """
        Call the table builder on one base case, as given in :self.current_base_case:
        :arg source_box_extent
        :arg alpha
        :arg delta
        :arg nlevels
        :arg extra_kernel_kwargs
        """

        if base_case_id != self.current_base_case:
            self.current_base_case = base_case_id

        if "source_box_extent" in kwargs:
            assert kwargs["source_box_extent"] > 0
            source_box_extent = kwargs["source_box_extent"]
        else:
            source_box_extent = 1

        extra_kernel_kwargs = {}
        if "extra_kernel_kwargs" in kwargs:
            extra_kernel_kwargs = kwargs["extra_kernel_kwargs"]

        if "alpha" in kwargs:
            alpha = kwargs["alpha"]
            assert alpha >= 0 and alpha < 1
        else:
            alpha = 0

        # (template) target points in 1D over [0, 1]
        q_points = self.get_target_points(queue)
        assert len(q_points) == self.ntgt_points ** self.dim
        t = np.array([pt[-1] for pt in q_points[: self.ntgt_points]])

        tt = t * source_box_extent
        delta_max = min(np.min(tt), source_box_extent - np.max(tt))
        assert delta_max > 0
        if delta_max < 1e-6:
            logger.warn("Severe delta constraint (< %f)" % delta_max)

        if ("delta" in kwargs) and (not self.auto_windowing):
            delta = kwargs["delta"]
            logger.info("Using window radius %f" % delta)
            assert delta > 0 and 2 * delta < source_box_extent
        else:
            assert self.auto_windowing
            delta = 0.9 * delta_max
            logger.info("Using auto-determined window radius %f" % delta)

        if delta > delta_max:
            delta = min(delta, delta_max)
            logger.info("Shrinked delta to %f to fit inside the source box" % delta)

        if "nlevels" in kwargs:
            nlevels = kwargs["nlevels"]
            assert nlevels > 0
        else:
            # Single level is equivalent to Duffy transform
            nlevels = 1

        missing_measure = (alpha ** nlevels * source_box_extent) ** self.dim
        if abs(missing_measure) > 1e-6:
            from warnings import warn

            warn(
                "Droste probably has too few levels, missing measure = "
                + str(missing_measure)
            )

        if "result_array" in kwargs:
            result_array = kwargs["result_array"]
        else:
            result_array = self.make_result_array(**kwargs)

        # root brick
        root_brick = np.zeros((self.dim, 2))
        root_brick[:, 1] = source_box_extent

        base_case_vec = np.array(
            [
                [self.reduce_by_symmetry.reduced_vecs[self.current_base_case][d]]
                for d in range(self.dim)
            ]
        )
        base_case_scl = np.array(
            [
                self.interaction_case_scls[
                    self.reduce_by_symmetry.reduced_vec_ids[self.current_base_case]
                ]
            ]
        )

        # quad formula for each brick (normalized to [0,1])
        # sps.legendre blows up easily at high order
        import scipy.special as sps

        # legendre_nodes, _, legendre_weights = sps.legendre(
        #        nquad_points).weights.T
        legendre_nodes, legendre_weights = sps.p_roots(self.nquad_points)
        legendre_nodes = legendre_nodes * 0.5 + 0.5
        legendre_weights = legendre_weights * 0.5

        # --------- call kernel 0 ----------
        self.get_kernel_id = 0
        try:
            delattr(self, "_memoize_dic_get_cached_optimized_kernel")
        except Exception:
            pass
        knl = self.get_cached_optimized_kernel()
        result_array_0 = self.make_result_array(**kwargs)
        evt0, res0 = knl(
            queue,
            alpha=alpha, delta=delta,
            result=result_array_0,
            root_brick=root_brick,
            target_nodes=t.astype(np.float64, copy=True),
            quadrature_nodes=legendre_nodes,
            quadrature_weights=legendre_weights,
            interaction_case_vecs=base_case_vec.astype(np.float64, copy=True),
            interaction_case_scls=base_case_scl.astype(np.float64, copy=True),
            n_cases=1,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )

        # --------- call kernel 1 ----------
        self.get_kernel_id = 1
        try:
            delattr(self, "_memoize_dic_get_cached_optimized_kernel")
        except Exception:
            pass
        result_array_1 = self.make_result_array(**kwargs)
        knl = self.get_cached_optimized_kernel()
        evt1, res1 = knl(
            queue,
            alpha=alpha, delta=delta,
            result=result_array_1,
            root_brick=root_brick,
            target_nodes=t.astype(np.float64, copy=True),
            quadrature_nodes=legendre_nodes,
            quadrature_weights=legendre_weights,
            interaction_case_vecs=base_case_vec.astype(np.float64, copy=True),
            interaction_case_scls=base_case_scl.astype(np.float64, copy=True),
            n_cases=1,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )

        # --------- call kernel 2 ----------
        self.get_kernel_id = 2
        try:
            delattr(self, "_memoize_dic_get_cached_optimized_kernel")
        except Exception:
            pass
        knl2 = self.get_cached_optimized_kernel()
        result_array = res0['result'] + res1['result']

        evt2, res2 = knl2(
            queue,
            result=result_array,
            n_cases=1,
            nfunctions=self.nfunctions,
            quad_order=self.ntgt_points,
            nlevels=nlevels,
            **extra_kernel_kwargs
        )

        cheb_table_case = res2["result"]
        return cheb_table_case

# }}} End inverse Droste method



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2019 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import pytest
from volumential.function_extension import compute_harmonic_extension
import numpy as np

import numpy.linalg as la
import pyopencl as cl
import pyopencl.clmath  # noqa

from meshmode.discretization import Discretization
from meshmode.discretization.poly_element import \
        InterpolatoryQuadratureSimplexGroupFactory

from pytential import bind, sym, norm  # noqa
from pytential.target import PointsTarget


def test_harmonic_extension_exterior_3d(ctx_factory):

    dim = 3  # noqa: F841
    mesh_order = 8
    fmm_order = 3
    bdry_quad_order = mesh_order
    bdry_ovsmp_quad_order = 4 * bdry_quad_order
    qbx_order = mesh_order

    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)

    from meshmode.mesh.generation import generate_icosphere
    from meshmode.mesh.refinement import refine_uniformly

    radius = 2.5
    base_mesh = generate_icosphere(radius, order=mesh_order)
    mesh = refine_uniformly(base_mesh, 1)

    pre_density_discr = Discretization(
            cl_ctx, mesh,
            InterpolatoryQuadratureSimplexGroupFactory(bdry_quad_order))

    from pytential.qbx import QBXLayerPotentialSource
    qbx, _ = QBXLayerPotentialSource(
            pre_density_discr, fine_order=bdry_ovsmp_quad_order, qbx_order=qbx_order,
            fmm_order=fmm_order,
            ).with_refinement()
    density_discr = qbx.density_discr

    ntgts = 200
    rho = np.random.rand(ntgts) * 3 + radius + 0.05
    theta = np.random.rand(ntgts) * 2 * np.pi
    phi = (np.random.rand(ntgts) - 0.5) * np.pi

    nodes = density_discr.nodes().with_queue(queue)
    source = np.array([0, 1, 2])

    def test_func(x):
        return 1.0/la.norm(x.get()-source[:, None], axis=0)

    f = cl.array.to_device(queue, test_func(nodes))

    targets = cl.array.to_device(queue,
            np.array([
                rho * np.cos(phi) * np.cos(theta),
                rho * np.cos(phi) * np.sin(theta),
                rho * np.sin(phi)])
            )

    exact_f = test_func(targets)

    ext_f, _ = compute_harmonic_extension(
            queue, PointsTarget(targets), qbx, density_discr,
            f, loc_sign=1)

    assert np.linalg.norm(exact_f - ext_f.get()) < 1e-3 * np.linalg.norm(exact_f)


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        pytest.main([__file__])


# vim: filetype=pyopencl:foldmethod=marker
